# -*- coding: utf-8 -*-

"""
@author: hannes.rotzinger@kit.edu / 2015,2016,2017 
         marco.pfirrmann@kit.edu / 2016, 2017
@license: GPL

"""
import sys
in_pyqt5 = False
try:
    from PyQt5 import QtCore
    import PyQt5.QtWidgets as QtGui
    in_pyqt5 = True
except ImportError as e:
    pass

if not in_pyqt5:
    try:
        from PyQt4 import QtCore, QtGui
    except ImportError:
        print("import of PyQt5 and PyQt4 failed. Install one of those.")
        sys.exit(-1)

import qkit
from qkit.storage.hdf_constants import ds_types

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)

class Ui_Form(object):
    """Ui_form class builds the general UI for the plot windows.

    This class creates the overall plot window based on the QtGui class. 
    Depending on the ds_type we add some signal slots for the user to add more
    possiblilities to view the data.
    setupUi() creates the overall window and ds_type sensitive signal slots 
    are added by the respective functions.
    """
    def setupUi(self, Form,ds_type, selector_labels):
        """Sets up the general window
        
        This function coordinates the changing input from the signal slots and
        updates the attributes that are parsed to the plotting lib functions.
        update_plots() is either periodically called e.g. by the timer or once 
        on startup.
        
        Args:
            self: Object of the Ui_Form class.
            Form: PlotWindow object that inherits the used calls here from the
                underlying QWidget class.
            ds_type: Integer
            selector_labels: String list with names of the datasets on all axis
        Returns:
            No return variable. The function operates on the given object.
        """
        Form.setObjectName(_fromUtf8("Form"))
        Form.resize(750, 450)
        self.gridLayout_Top = QtGui.QGridLayout(Form)
        self.gridLayout_Top.setContentsMargins(0,0,0,0)
        self.gridLayout_Top.setObjectName(_fromUtf8("gridLayout_Top"))
        # we push for a tight layout
        self.gridLayout_Top.setContentsMargins(0,0,0,0)
        self.gridLayout_Top.setContentsMargins(QtCore.QMargins(0,0,0,0))
        self.gridLayout_Top.setSpacing(0)
        
        self.horizontalLayout = QtGui.QHBoxLayout()
        self.horizontalLayout.setSizeConstraint(QtGui.QLayout.SetMinimumSize)
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))

        self.horizontalLayout.setContentsMargins(0,0,0,0)
        self.horizontalLayout.setContentsMargins(QtCore.QMargins(0,0,0,0))
        self.horizontalLayout.setSpacing(0)
        

        self.gridLayout_Top.addLayout(self.horizontalLayout, 0, 0, 1, 1)
        self.gridLayout = QtGui.QGridLayout()
        self.gridLayout.setObjectName(_fromUtf8("gridLayout"))
        
        self.gridLayout_Top.addLayout(self.gridLayout, 1, 0, 1, 1)

        Form.setWindowTitle(_translate("Form", "Form", None))
        QtCore.QMetaObject.connectSlotsByName(Form)        
        
        self.selector_labels = selector_labels
        if ds_type == ds_types['coordinate']:
            self.setupCoordinate(Form)
        if ds_type == ds_types['vector']:
            self.setupVector(Form)
        if ds_type == ds_types['matrix'] or ds_type == -1:
            self.setupMatrix(Form)
        if ds_type == ds_types['box']:
            self.setupBox(Form)
        if ds_type == ds_types['txt']:
            self.setupTxt(Form)
        if ds_type == ds_types['view']:
            self.setupView(Form)
        
    def setupCoordinate(self,Form):
        # see setupVector()
        self.setupVector(Form)

    def setupVector(self,Form):
        """Set up slots for value_vector.
        
        It is possible to plot the data over the x_axis and display the numeric
        values in a table.
        Args:
            self: Object of the Ui_Form class.
            Form: PlotWindow object that inherits the used calls here from the
                underlying QWidget class.
        Returns:
            No return variable. The function operates on the given object.
        """
        self.PlotTypeSelector = QtGui.QComboBox(Form)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.PlotTypeSelector.sizePolicy().hasHeightForWidth())
        self.PlotTypeSelector.setSizePolicy(sizePolicy)
        self.PlotTypeSelector.setObjectName(_fromUtf8("PlotTypeSelector"))
        self.PlotTypeSelector.addItem(_fromUtf8(""))
        self.PlotTypeSelector.addItem(_fromUtf8(""))
        self.horizontalLayout.addWidget(self.PlotTypeSelector)
        self.PlotTypeSelector.setItemText(0, _translate("Form", "Line Plot", None))
        self.PlotTypeSelector.setItemText(1, _translate("Form", "Table", None))
        
        spacerItem = QtGui.QSpacerItem(40, 1, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self._addIndicatorLabels(Form,sizePolicy,indicators=["PointX","PointY"])
        

    def setupView(self,Form):
        """Set up slots for views.
        
        The x- and y-values of higher dimensional datasets can be selected to
        be displayed in the 1d plot.

        Args:
            self: Object of the Ui_Form class.
            Form: PlotWindow object that inherits the used calls here from the
                underlying QWidget class.
        Returns:
            No return variable. The function operates on the given object.
        """
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        self._addTraceSelectorIndicator(Form,sizePolicy,TraceSelector = "VTraceXSelector", 
                                        TraceIndicator="VTraceXValue", prefix = self.selector_labels[3]+': ')
        self._addTraceSelectorIndicator(Form,sizePolicy,TraceSelector = "VTraceYSelector", 
                                        TraceIndicator="VTraceYValue", prefix = self.selector_labels[4]+': ')        

        spacerItem = QtGui.QSpacerItem(40, 1, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        
        self._addIndicatorLabels(Form,sizePolicy,indicators=["PointX","PointY"])
        

    def setupMatrix(self,Form):
        """Set up slots for value_matrix.
        
        The data can be plotted color coded as a 2d plot, as a 1d plot at a
        user selected value of the x_axis or the numerical values can be
        displayed in a table.

        Args:
            self: Object of the Ui_Form class.
            Form: PlotWindow object that inherits the used calls here from the
                underlying QWidget class.
        Returns:
            No return variable. The function operates on the given object.
        """
        self.PlotTypeSelector = QtGui.QComboBox(Form)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Preferred, QtGui.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.PlotTypeSelector.sizePolicy().hasHeightForWidth())
        
        self.PlotTypeSelector.setSizePolicy(sizePolicy)
        self.PlotTypeSelector.setObjectName(_fromUtf8("PlotTypeSelector"))
        self.PlotTypeSelector.addItem(_fromUtf8(""))
        self.PlotTypeSelector.addItem(_fromUtf8(""))
        self.PlotTypeSelector.addItem(_fromUtf8(""))
        self.PlotTypeSelector.addItem(_fromUtf8(""))
        self.PlotTypeSelector.setItemText(0, _translate("Form", "Color Plot", None))
        self.PlotTypeSelector.setItemText(1, _translate("Form", "Line Plot X", None))
        self.PlotTypeSelector.setItemText(2, _translate("Form", "Line Plot Y", None))
        self.PlotTypeSelector.setItemText(3, _translate("Form", "Table", None))
        
        
        self.PlotTypeLayout = QtGui.QVBoxLayout()
        self.PlotTypeLayout.addWidget(self.PlotTypeSelector)
        # add a empty label to move the PlotTypeSelector to the top
        emptyL = QtGui.QLabel(Form)
        self.PlotTypeLayout.addWidget(emptyL)
        self.horizontalLayout.addLayout(self.PlotTypeLayout,stretch = -10)
        
        self._addTraceSelectorIndicator(Form,sizePolicy,TraceSelector = "TraceXSelector",
                                        TraceIndicator="TraceXValue", prefix = self.selector_labels[0]+': ')
        self._addTraceSelectorIndicator(Form,sizePolicy,TraceSelector = "TraceYSelector", 
                                        TraceIndicator="TraceYValue", prefix = self.selector_labels[1]+': ')
        
        #The indicators should be located at the most right side of the bar
        spacerItem = QtGui.QSpacerItem(40, 1, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)        
        self._addIndicatorLabels(Form,sizePolicy, indicators=["PointX","PointY","PointZ"])
        
        
    def setupBox(self,Form):
        """Set up slots for value_box.
        
        The data can be plotted color coded as a 2d plot at a user selected
        value of either the x_, y_ or z_axis or as a 1d plot at a user selected 
        value of the x_ and y_axis.

        Args:
            self: Object of the Ui_Form class.
            Form: PlotWindow object that inherits the used calls here from the
                underlying QWidget class.
        Returns:
            No return variable. The function operates on the given object.
        """
        self.PlotTypeSelector = QtGui.QComboBox(Form)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.PlotTypeSelector.sizePolicy().hasHeightForWidth())
        
        self.PlotTypeSelector.setSizePolicy(sizePolicy)
        self.PlotTypeSelector.setObjectName(_fromUtf8("PlotTypeSelector"))
        self.PlotTypeSelector.addItem(_fromUtf8(""))
        self.PlotTypeSelector.addItem(_fromUtf8(""))
        self.PlotTypeSelector.addItem(_fromUtf8(""))
        self.PlotTypeSelector.addItem(_fromUtf8(""))
        self.PlotTypeSelector.addItem(_fromUtf8(""))
        self.PlotTypeSelector.addItem(_fromUtf8(""))
        #self.horizontalLayout.addWidget(self.PlotTypeSelector)
        self.PlotTypeSelector.setItemText(0, _translate("Form", "2D Select X", None))
        self.PlotTypeSelector.setItemText(1, _translate("Form", "2D Select Y", None))
        self.PlotTypeSelector.setItemText(2, _translate("Form", "2D Select Z", None))
        self.PlotTypeSelector.setItemText(3, _translate("Form", "Line Plot X", None))
        self.PlotTypeSelector.setItemText(4, _translate("Form", "Line Plot Y", None))
        self.PlotTypeSelector.setItemText(5, _translate("Form", "Line Plot Z", None))
        
        self.PlotTypeLayout = QtGui.QVBoxLayout()
        self.PlotTypeLayout.addWidget(self.PlotTypeSelector)
        # add a empty label to move the PlotTypeSelector to the top
        emptyL = QtGui.QLabel(Form)
        self.PlotTypeLayout.addWidget(emptyL)
        self.horizontalLayout.addLayout(self.PlotTypeLayout,stretch = -10)
       
        self._addTraceSelectorIndicator(Form,sizePolicy,TraceSelector = "TraceXSelector", 
                                        TraceIndicator="TraceXValue", prefix = self.selector_labels[0]+': ')
        self._addTraceSelectorIndicator(Form,sizePolicy,TraceSelector = "TraceYSelector", 
                                        TraceIndicator="TraceYValue", prefix = self.selector_labels[1]+': ')
        self._addTraceSelectorIndicator(Form,sizePolicy,TraceSelector = "TraceZSelector",  
                                        TraceIndicator="TraceZValue",  prefix = self.selector_labels[2]+': ')        

        spacerItem = QtGui.QSpacerItem(40, 1, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        
        self._addIndicatorLabels(Form,sizePolicy,indicators=["PointX","PointY","PointZ"])
        

    def setupTxt(self,Form):
        pass

    def _addIndicatorLabels(self,Form,sizePolicy,indicators=[]):
        self.IndicatorLayout = QtGui.QVBoxLayout()
        self.IndicatorLayout.setSizeConstraint(QtGui.QLayout.SetMinimumSize)
        self.IndicatorLayout.setObjectName(_fromUtf8("horizontalLayout"))

        self.IndicatorLayout.setContentsMargins(0,0,0,0)
        self.IndicatorLayout.setContentsMargins(QtCore.QMargins(0,0,0,0))
        self.IndicatorLayout.setSpacing(3)
        
        for indicator in indicators:
            setattr(self,indicator,QtGui.QLabel(Form))
            temp_indicator = getattr(self,indicator)
            temp_indicator.setSizePolicy(sizePolicy)
            temp_indicator.setObjectName(_fromUtf8(indicator))
            self.IndicatorLayout.addWidget(temp_indicator)
            
        self.horizontalLayout.addLayout(self.IndicatorLayout,stretch = -10)


    def _addTraceSelectorIndicator(self,Form,sizePolicy,TraceSelector = "", TraceIndicator="", prefix = "Trace: "):
        self.TraceSelIndLayout = QtGui.QVBoxLayout()
        self.TraceSelIndLayout.setSizeConstraint(QtGui.QLayout.SetMinimumSize)
        self.TraceSelIndLayout.setObjectName(_fromUtf8("TraceSelIndLayout"))

        self.TraceSelIndLayout.setContentsMargins(0,0,0,0)
        self.TraceSelIndLayout.setContentsMargins(QtCore.QMargins(0,0,0,0))
        self.TraceSelIndLayout.setSpacing(1)


        setattr(self,TraceSelector,QtGui.QSpinBox(Form))
        temp_SelInd = getattr(self,TraceSelector)
        temp_SelInd.setSizePolicy(sizePolicy)
        temp_SelInd.setSuffix(_fromUtf8(""))
        temp_SelInd.setMinimum(-99999)
        temp_SelInd.setMaximum(99999)
        temp_SelInd.setProperty("value", -1)
        temp_SelInd.setObjectName(_fromUtf8(TraceSelector))
        temp_SelInd.setPrefix(_translate("Form", prefix , None))

        
        self.TraceSelIndLayout.addWidget(temp_SelInd)
            
        setattr(self,TraceIndicator,QtGui.QLineEdit(Form))
        temp_SelInd = getattr(self,TraceIndicator)
        temp_SelInd.setSizePolicy(sizePolicy)
        temp_SelInd.setReadOnly(False)
        temp_SelInd.setObjectName(_fromUtf8("TraceValue"))
        self.TraceSelIndLayout.addWidget(temp_SelInd)
        self.horizontalLayout.addLayout(self.TraceSelIndLayout,stretch = -10)              


# -*- coding: utf-8 -*-
"""
Created 2015

@author: hrotzing
"""
import qkit
from qkit.storage.hdf_constants import ds_types, view_types
import json
from qkit.core.lib.misc import str3
class dataset_view(object):
    """This class describes a view on one or two datasets.        
    
    Views do not contain any data but only ds_url information about the datasets
    that should  be displayed. The attributes 'overlays' and 'xy_i' hold the 
    information about how many plots are created over one another and which
    datasets are plotted against each other.
    """
    
    def __init__(self, hdf_file, name, x=None, y=None, error=None, filter = None,
                 ds_type =  ds_types['view'], folder = 'views', view_params={}):
        
        self.hf = hdf_file
        self.name = name
        self.folder = folder
        self.ds_url = "/entry/" + folder + "/" + name
        self.ds_type = ds_type
        self.filter = filter
        if not x or not y:
            raise ValueError(__name__ + ": Error creating view '{!s}' in file {!s}: x or y axis not given.".format(name, hdf_file.hf.attrs['_filepath']))
        self.x_object = str3(x.ds_url)
        self.y_object = str3(y.ds_url)
        self.error_object = None
        if error:
            self.error_object = str3(error.ds_url)
        
        self.ds = self.hf.create_dataset(self.name,0,folder=self.folder,dim=1)
        self.view_params = json.dumps(view_params)
        self.view_num = 0
        self._setup_metadata()
        self.hf.flush()

        
    def add(self,x=None, y=None, error=None, pyfilter = None):
        """Adds one y vs x overlay plot to the dataset_view object.        
    
        Args:
            x, y, error: hdf_datasets; y over x gets plotted with error being 
                the y-error data.
        """
        if x:
            self.x_object = str3(x.ds_url)
        self.y_object = str3(y.ds_url)
        self.error_object = None
        if error:
            self.error_object = str3(error.ds_url)
        self.filter = pyfilter
        self._setup_metadata(init = False)
        
    def _setup_metadata(self,init=True):
        ds = self.ds
        if init:
            ds.attrs.create("view_type",view_types['1D-V'])
            ds.attrs.create('ds_type',self.ds_type)
            ds.attrs.create('view_params',self.view_params.encode())
        ds.attrs.create("xy_"+str(self.view_num),(str(self.x_object)+":"+str(self.y_object)).encode())
        if self.error_object:
            ds.attrs.create("xy_"+str(self.view_num)+"_error",str(self.error_object).encode())
        ds.attrs.create("overlays",self.view_num)
        ds.attrs.create("xy_"+str(self.view_num)+"_filter",str(self.filter).encode())

        self.view_num += 1
        
    def _exec_filter(self,view_num):
        """
        FIXME: this does not work!
        # this is a somewhat dangerous function, since it allows to modify the 
        # hdf file while it is being updated by another program
        # as usual, power comes with responsibility...
        
        # prepare filter
        x_ds_url,y_ds_url = self.ds.attrs.get("xy_" + str(view_num)).split(':')
        print x_ds_url,y_ds_url
        
        locs = { 'x' : x_ds_url,
                  'y' : y_ds_url }
        head = "from qkit.config.environment import *\n"
        flt = head + self.filter
        
        try:
            exec(flt, globals(), locs)
            #exec(self.filter,locs)
        except SyntaxError as e:
            print("The filter code has to be proper python code! %s "% e)
        """
        pass



import qkit
from qkit.core.lib import temp

#qt.flow.register_exit_handler(qt.config._do_save)
#qt.flow.register_exit_handler(qt.flow.close_gui)

qkit.flow.register_exit_handler(temp.File.remove_all)

# Clear "starting" status
qkit.flow.finished_starting()



# QKIT driver for remote access of an microwave switch control server
# Author: S1 @ KIT 2020

import zmq

from qkit.core.instrument_base import Instrument


class switch_client(Instrument):
    """
        This is the remote switch client to connect to the v2 switch raspberry

        Usage:
        Initialize with
        <name> = instruments.create('<name>', 'switch_client', url = "tcp://localhost:5000")
    """
    
    def __init__(self, name, url="tcp://localhost:5000"):
        
        Instrument.__init__(self, name, tags=['physical'])
        
        # Add some global constants
        # self._address = address
        # self._port = port
        
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.REQ)
        self.socket.setsockopt(zmq.LINGER, 0)
        self.setup_connection(url=url)
        self.default_device = ''
        self.control_device = ''
        
        self.add_function('set_pulse_length')
        self.add_function('get_pulse_length')
        self.add_function('enable')
        self.add_function('disable')
        self.add_function('switch_to')
        self.add_function('is_at_position')
        self.add_function('ensure_switch_at')
    
    def close(self):
        print("closing zmq socket")
        self.socket.close()
    
    def setup_connection(self, url="tcp://localhost:5000"):
        print("Connecting to switch server...")
        self.socket.connect(url)
    
    def enable(self, switch, port):
        self.socket.send_string("enable/%i/%i" % (switch, port))
        return self.socket.recv_string()
    
    def disable(self, switch, port):
        self.socket.send_string("disable/%i/%i" % (switch, port))
        return self.socket.recv_string()
    
    def switch_to(self, switch, port):
        self.socket.send_string("set/%i/%i" % (switch, port))
        return self.socket.recv_string()
    
    def set_pulse_length(self, pulse_length, switch=None):
        if switch is None:
            self.socket.send_string("set/length/%.3f" % pulse_length)
        else:
            self.socket.send_string("set/length/%.3f/%i" % (pulse_length, switch))
        return self.socket.recv_string()
    
    def get_pulse_length(self):
        self.socket.send_string("get/length")
        return self.socket.recv_string()
    
    def reset(self, switch):
        self.socket.send_string("reset/%i" % switch)
        return self.socket.recv_string()
    
    def get_position(self,switch):
        self.socket.send_string("get/%i" % switch)
        return self.socket.recv_string()
    
    def is_at_position(self, switch, port):
        self.socket.send_string("get/%i/%i" % (switch, port))
        response = self.socket.recv_string() # This is a string of either "1" or "0", as per server code
        return bool(int(response)) # convert string ("1"|"0") to in (1|0) to bool (True|False)
    
    def ensure_switch_at(self, switch, port):
        """
        Ensures that the switch is at the specified position.
        Idempotent.
        """
        if not self.is_at_position(switch, port):
            self.switch_to(switch, port)



# MP@KIT 2015
# HR@KIT 2015
#import h5py
import numpy as np
import logging

import qkit
from qkit.storage import store
from qkit.analysis.circle_fit import circuit
from qkit.storage.hdf_constants import ds_types
from scipy.optimize import leastsq
from scipy.ndimage import gaussian_filter1d
from scipy.ndimage.filters import median_filter

class Resonator(object):
    '''
    Resonator class for fitting (live or after measurement) amplitude and phase data at multiple functions. The data is stored in .h5-files, having a NeXus compatible organization.
    Required are datasets frequency (/entry/data0/frequency), amplitude (/entry/data0/amplitude), and phase (/entry/data0/phase).
    input:
    hf_path (HDF5-filepath): Path to file containing datasets to be fitted

    Possible fits are 'lorentzian', 'skewed lorentzian', 'circle', and 'fano' with each fit taking arguments
    fit_all (boolean, optional): True or False, default: False, fit all entries in the amplitude dataset or only last one
    f_min (float, optional): Lower boundary for fit function
    f_max (float, optional): Upper boundary for fit function

    usage:
        res=Resonator(filepath)
        res.fit_lorentzian(fit_all=True,f_min=5.667e9,f_max=5.668e9)
        res.fit_fano(fit_all=True)
        res.fit_circle(fit_all=True,f_max=5.668e9)
    '''

    def __init__(self, hf_path):
        self._hf = store.Data(hf_path)

        self._first_circle = True
        self._first_lorentzian = True
        self._first_fano = True
        self._first_skewed_lorentzian = True
        
        self._do_prefilter_data = False
        self.pre_filter_params = []
        self._debug = False

        # these ds_url should always be present in a resonator measurement
        self.ds_url_amp = "/entry/data0/amplitude"
        self.ds_url_pha = "/entry/data0/phase"
        self.ds_url_freq = "/entry/data0/frequency"

        # these ds_url depend on the measurement and may not exist
        self.ds_url_power = "/entry/data0/power"
        
        
        '''
        catching error if datasets are empty while creating resonator object
        i.g. while live-fitting
        '''
        try:
            self._get_datasets()
            self._datasets_loaded = True
        except KeyError:
            self._datasets_loaded = False

    def debug(self,message):
        if self._debug:
            print(message)

    def set_file(self,hf):
        '''
        sets hf file
        input:
        hf (HDF5-file)
        '''
        self._hf=hf
        self._prepare()

    def close(self):
        self._hf.close()

    def set_x_coord(self,x_co):
        '''
        sets x-coordinate for the datasets
        input:
        x_co (string): url of the x-coordinate
        '''
        try:
            self._x_co = self._hf.get_dataset(x_co)
        except:
            logging.warning('Unable to open any x_coordinate. Please set manually using \'set_x_coord()\'.')

    def set_y_coord(self,y_co):
        '''
        sets y-coordinate for the datasets
        input:
        y_co (string): url of the y-coordinate
        '''
        try:
            self._y_co = self._hf.get_dataset(y_co)
        except:
            logging.warning('Unable to open any y_coordinate. Please set manually using \'set_y_coord()\'.')

    def set_prefilter(self,gaussian = False, median = False, params = []):
        self._do_prefilter_data = False
        if gaussian or median:
            self._do_prefilter_data = True

        #print gaussian, median
        if median:
            self._prefilter = median_filter
            #print("median_filter")
            if params:
                self._prefilter_params = params[0]
            else:
                self._prefilter_params = 6

        if gaussian:
            self._prefilter = gaussian_filter1d
            #print("gaussian_filter1d")
            if params:
                self._prefilter_params = params[0]
            else:
                self._prefilter_params = 6 # 0.4

    def _pre_filter_data(self,data):
        # in the moment only gaussian and median are suported
        if self._do_prefilter_data:
                return self._prefilter(data,self._prefilter_params)
        else:
            return data

    def _set_data_range(self, data):
        '''
        cuts the data array to the positions where f>=f_min and f<=f_max in the frequency-array
        the fit functions are fitted only in this area
        the data in the .h5-file is NOT changed
        '''
        if data.ndim == 1:
            return data[(self._frequency >= self._f_min) & (self._frequency <= self._f_max)]
        if data.ndim == 2:
            ret_array=np.empty(shape=(data.shape[0],self._fit_frequency.shape[0]),dtype=np.float64)
            for i,a in enumerate(data):
                ret_array[i]=data[i][(self._frequency >= self._f_min) & (self._frequency <= self._f_max)]
            return ret_array

    def _get_datasets(self):
        '''
        reads out the file
        '''
        if not self._hf:
            logging.info('No hf file kown yet!')
            return

        self._ds_amp = self._hf.get_dataset(self.ds_url_amp)
        self._ds_pha = self._hf.get_dataset(self.ds_url_pha)
        self._ds_type = self._ds_amp.ds_type

        self._amplitude = np.array(self._hf[self.ds_url_amp],dtype=np.float64)
        self._phase = np.array(self._hf[self.ds_url_pha],dtype=np.float64)
        self._frequency = np.array(self._hf[self.ds_url_freq],dtype=np.float64)

        try:
            self._x_co = self._hf.get_dataset(self._ds_amp.x_ds_url)
        except:
            try:
                self._x_co = self._hf.get_dataset(self.ds_url_power) # hardcode a std url
            except:
                logging.warning('Unable to open any x_coordinate. Please set manually using \'set_x_coord()\'.')
        try:
            self._y_co = self._hf.get_dataset(self._ds_amp.y_ds_url)
        except:
            try: self._y_co = self._hf.get_dataset(self.ds_url_freq) # hardcode a std url
            except:
                logging.warning('Unable to open any y_coordinate. Please set manually using \'set_y_coord()\'.')
        self._datasets_loaded = True

    def _prepare_f_range(self,f_min,f_max):
        '''
        prepares the data to be fitted:
        f_min (float): lower boundary
        f_max (float): upper boundary
        '''

        self._f_min = np.min(self._frequency)
        self._f_max = np.max(self._frequency)

        '''
        f_min f_max do not have to be exactly an entry in the freq-array
        '''
        if f_min:
            for freq in self._frequency:
                if freq > f_min:
                    self._f_min = freq
                    break
        if f_max:
            for freq in self._frequency:
                if freq > f_max:
                    self._f_max = freq
                    break

        '''
        cut the data-arrays with f_min/f_max and fit_all information
        '''
        self._fit_frequency = np.array(self._set_data_range(self._frequency))
        self._fit_amplitude = np.array(self._set_data_range(self._amplitude))
        self._fit_phase = np.array(self._set_data_range(self._phase))

        self._frequency_co = self._hf.add_coordinate('frequency',folder='analysis', unit = 'Hz')
        self._frequency_co.add(self._fit_frequency)

    def _update_data(self):
        self._amplitude = np.array(self._hf[self.ds_url_amp],dtype=np.float64)
        self._phase = np.array(self._hf[self.ds_url_pha],dtype=np.float64)

    def _get_starting_values(self):
        pass
    
    def fit_circle(self,reflection = False, notch = False, fit_all = False, f_min = None, f_max=None):
        self._fit_all = fit_all
        self._circle_reflection = reflection
        self._circle_notch = notch
        if not reflection and not notch:
            self._circle_notch = True
        
        if not self._datasets_loaded:
            self._get_datasets()

        self._update_data()
        self._prepare_f_range(f_min, f_max)
        
        if self._first_circle:
            self._prepare_circle()
            self._first_circle = False

        if self._circle_reflection:
           self._circle_port = circuit.reflection_port(f_data = self._fit_frequency)
        elif self._circle_notch:
            self._circle_port = circuit.notch_port(f_data = self._fit_frequency)
            
        self._do_fit_circle()

    def _do_fit_circle(self):
        '''
        Creates corresponding ports in circuit.py in the qkit/analysis folder
        circle fit for amp and pha data in the f_min-f_max frequency range
        fit parameter, errors, and generated amp/pha data are stored in the hdf-file

        input:
        fit_all (bool): True or False, default: False. Whole data (True) or only last "slice" (False) is fitted (optional)
        '''

        self._get_data_circle()
        trace = 0
        self.debug("circle fit:")
        for z_data_raw in self._z_data_raw:
            
            z_data_raw.real = self._pre_filter_data(z_data_raw.real)
            z_data_raw.imag = self._pre_filter_data(z_data_raw.imag)
            self.debug("fitting trace: "+str(trace))
            self._circle_port.z_data_raw = z_data_raw
            
            try:
                self._circle_port.autofit()
            except:
                err=np.array([np.nan for f in self._fit_frequency])
                self._circ_amp_gen.append(err)
                self._circ_pha_gen.append(err)
                self._circ_real_gen.append(err)
                self._circ_imag_gen.append(err)

                for key in iter(self._results):
                    self._results[str(key)].append(np.nan)

            else:
                self._circ_amp_gen.append(np.absolute(self._circle_port.z_data_sim))
                self._circ_pha_gen.append(np.angle(self._circle_port.z_data_sim))
                self._circ_real_gen.append(np.real(self._circle_port.z_data_sim))
                self._circ_imag_gen.append(np.imag(self._circle_port.z_data_sim))

                for key in iter(self._results):
                    self._results[str(key)].append(float(self._circle_port.fitresults[str(key)]))
            trace+=1

    def _prepare_circle(self):
        '''
        creates the datasets for the circle fit in the hdf-file
        '''
        self._results = {}
        circle_fit_version = qkit.cfg.get("circle_fit_version", 1)
        if circle_fit_version == 1:
            if self._circle_notch:
                self._result_keys = ["Qi_dia_corr", "Qi_no_corr", "absQc", "Qc_dia_corr", "Ql",
                                     "fr", "theta0", "phi0", "phi0_err", "Ql_err", "absQc_err",
                                     "fr_err", "chi_square", "Qi_no_corr_err", "Qi_dia_corr_err"]
            elif self._circle_reflection:
                self._result_keys = ["Qi", "Qc", "Ql", "fr", "theta0", "Ql_err",
                                     "Qc_err", "fr_err", "chi_square", "Qi_err"]
                
        elif circle_fit_version == 2:
            self._result_keys = ["delay", "delay_remaining", "a", "alpha", "theta", "phi", "fr", "Ql", "Qc",
                                 "Qc_no_dia_corr", "Qi", "Qi_no_dia_corr", "fr_err", "Ql_err", "absQc_err",
                                 "phi_err", "Qi_err", "Qi_no_dia_corr_err", "chi_square", "Qi_min", "Qi_max",
                                 "Qc_min", "Qc_max", "fano_b"]
        else:
            logging.warning("Circle fit version not properly set in configuration!")

        if self._ds_type == ds_types['vector']:  # data from measure_1d
            self._data_real_gen = self._hf.add_value_vector('data_real_gen', self._frequency_co, folder='analysis',
                                                            unit='')
            self._data_imag_gen = self._hf.add_value_vector('data_imag_gen', self._frequency_co, folder='analysis',
                                                            unit='')

            self._circ_amp_gen = self._hf.add_value_vector('circ_amp_gen', self._frequency_co, folder='analysis',
                                                           unit='arb. unit')
            self._circ_pha_gen = self._hf.add_value_vector('circ_pha_gen', self._frequency_co, folder='analysis',
                                                           unit='rad')
            self._circ_real_gen = self._hf.add_value_vector('circ_real_gen', self._frequency_co, folder='analysis',
                                                            unit='')
            self._circ_imag_gen = self._hf.add_value_vector('circ_imag_gen', self._frequency_co, folder='analysis',
                                                            unit='')

            for key in self._result_keys:
                self._results[key] = self._hf.add_coordinate('circ_' + key, folder='analysis', unit='')

        if self._ds_type == ds_types['matrix']:  # data from measure_2d
            self._data_real_gen = self._hf.add_value_matrix('data_real_gen', self._x_co, self._frequency_co,
                                                            folder='analysis', unit='')
            self._data_imag_gen = self._hf.add_value_matrix('data_imag_gen', self._x_co, self._frequency_co,
                                                            folder='analysis', unit='')

            self._circ_amp_gen = self._hf.add_value_matrix('circ_amp_gen', self._x_co, self._frequency_co,
                                                           folder='analysis', unit='arb. unit')
            self._circ_pha_gen = self._hf.add_value_matrix('circ_pha_gen', self._x_co, self._frequency_co,
                                                           folder='analysis', unit='rad')
            self._circ_real_gen = self._hf.add_value_matrix('circ_real_gen', self._x_co, self._frequency_co,
                                                            folder='analysis', unit='')
            self._circ_imag_gen = self._hf.add_value_matrix('circ_imag_gen', self._x_co, self._frequency_co,
                                                            folder='analysis', unit='')

            for key in self._result_keys:
                self._results[key] = self._hf.add_value_vector('circ_' + key, folder='analysis', x=self._x_co, unit='')

        circ_view_amp = self._hf.add_view('circ_amp', x=self._y_co, y=self._ds_amp)
        circ_view_amp.add(x=self._frequency_co, y=self._circ_amp_gen)
        circ_view_pha = self._hf.add_view('circ_pha', x=self._y_co, y=self._ds_pha)
        circ_view_pha.add(x=self._frequency_co, y=self._circ_pha_gen)
        circ_view_iq = self._hf.add_view('circ_IQ', x=self._data_real_gen, y=self._data_imag_gen,
                                         view_params={'aspect': 1.0})
        circ_view_iq.add(x=self._circ_real_gen, y=self._circ_imag_gen)

    def _get_data_circle(self):
        '''
        calc complex data from amp and pha
        '''
        if not self._fit_all:
            self._z_data_raw = np.empty((1,self._fit_frequency.shape[0]), dtype=np.complex64)

            if self._fit_amplitude.ndim == 1: 
                self._z_data_raw[0] = np.array(self._fit_amplitude*np.exp(1j*self._fit_phase),dtype=np.complex64)
            else: 
                self._z_data_raw[0] = np.array(self._fit_amplitude[-1]*np.exp(1j*self._fit_phase[-1]),dtype=np.complex64)

            self._data_real_gen.append(self._z_data_raw[0].real)
            self._data_imag_gen.append(self._z_data_raw[0].imag)

        if self._fit_all:
            self._z_data_raw = np.empty((self._fit_amplitude.shape), dtype=np.complex64)
            for i,a in enumerate(self._fit_amplitude):
                self._z_data_raw[i] = self._fit_amplitude[i]*np.exp(1j*self._fit_phase[i])
                self._data_real_gen.append(self._z_data_raw[i].real)
                self._data_imag_gen.append(self._z_data_raw[i].imag)

    def _get_last_amp_trace(self):
        tmp_amp = np.empty((1,self._fit_frequency.shape[0]))
        if self._fit_amplitude.ndim==1: 
            tmp_amp[0] = self._fit_amplitude
        else:
            tmp_amp[0] = self._fit_amplitude[-1]
        self._fit_amplitude = np.empty((1,self._fit_frequency.shape[0]))
        self._fit_amplitude[0] = tmp_amp[0]

    def fit_lorentzian(self,fit_all = False,f_min=None,f_max=None,pre_filter_data=None):
        '''
        lorentzian fit for amp data in the f_min-f_max frequency range
        squared amps are fitted at lorentzian using scipy.leastsq
        fit parameter, chi2, and generated amp are stored in the hdf-file

        input:
        fit_all (bool): True or False, default: False. Whole data (True) or only last "slice" (False) is fitted (optional)
        f_min (float): lower boundary for data to be fitted (optional, default: None, results in min(frequency-array))
        f_max (float): upper boundary for data to be fitted (optional, default: None, results in max(frequency-array))
        '''
        def residuals(p,x,y):
            f0,k,a,offs=p
            err = y-(a/(1+4*((x-f0)/k)**2)+offs)
            return err
        
        self._fit_all = fit_all

        if not self._datasets_loaded:
            self._get_datasets()

        self._update_data()
        self._prepare_f_range(f_min,f_max)
        if self._first_lorentzian:
            self._prepare_lorentzian()
            self._first_lorentzian=False

        '''
        fit_amplitude is always 2dim np array.
        for 1dim data, shape: (1, # fit frequency points)
        for 2dim data, shape: (# amplitude slices (i.e. power values in scan), # fit frequency points)
        '''
        if not self._fit_all:
            self._get_last_amp_trace()

        for amplitudes in self._fit_amplitude:
            amplitudes = np.absolute(amplitudes)
            amplitudes_sq = amplitudes**2
            '''extract starting parameter for lorentzian from data'''
            s_offs = np.mean(np.array([amplitudes_sq[:int(np.size(amplitudes_sq)*.1)], amplitudes_sq[int(np.size(amplitudes_sq)-int(np.size(amplitudes_sq)*.1)):]]))
            '''offset is calculated from the first and last 10% of the data to improve fitting on tight windows'''

            if np.abs(np.max(amplitudes_sq)-np.mean(amplitudes_sq)) > np.abs(np.min(amplitudes_sq)-np.mean(amplitudes_sq)):
                '''peak is expected'''
                s_a = np.abs((np.max(amplitudes_sq)-np.mean(amplitudes_sq)))
                s_f0 = self._fit_frequency[np.argmax(amplitudes_sq)]
            else:
                '''dip is expected'''
                s_a = -np.abs((np.min(amplitudes_sq)-np.mean(amplitudes_sq)))
                s_f0 = self._fit_frequency[np.argmin(amplitudes_sq)]

            '''estimate peak/dip width'''
            mid = s_offs + .5*s_a #estimated mid region between base line and peak/dip
            m = [] #mid points
            for i in range(len(amplitudes_sq)-1):
                if np.sign(amplitudes_sq[i]-mid) != np.sign(amplitudes_sq[i+1]-mid):#mid level crossing
                    m.append(i)
            if len(m)>1:
                s_k = self._fit_frequency[m[-1]]-self._fit_frequency[m[0]]
            else:
                s_k = .15*(self._fit_frequency[-1]-self._fit_frequency[0]) #try 15% of window
            p0=[s_f0, s_k, s_a, s_offs]
            try:
                fit = leastsq(residuals,p0,args=(self._fit_frequency,amplitudes_sq))
            except:
                self._lrnz_amp_gen.append(np.array([np.nan for f in self._fit_frequency]))
                self._lrnz_f0.append(np.nan)
                self._lrnz_k.append(np.nan)
                self._lrnz_a.append(np.nan)
                self._lrnz_offs.append(np.nan)
                self._lrnz_Ql.append(np.nan)
                self._lrnz_chi2_fit.append(np.nan)
            else:
                popt=fit[0]
                chi2 = self._lorentzian_fit_chi2(popt,amplitudes_sq)
                self._lrnz_amp_gen.append(np.sqrt(np.array(self._lorentzian_from_fit(popt))))
                self._lrnz_f0.append(float(popt[0]))
                self._lrnz_k.append(float(np.fabs(float(popt[1]))))
                self._lrnz_a.append(float(popt[2]))
                self._lrnz_offs.append(float(popt[3]))
                self._lrnz_Ql.append(float(float(popt[0])/np.fabs(float(popt[1]))))
                self._lrnz_chi2_fit.append(float(chi2))

    def _prepare_lorentzian(self):
        '''
        creates the datasets for the lorentzian fit in the hdf-file
        '''
        if self._ds_type == ds_types['vector']: # data from measure_1d
            self._lrnz_amp_gen = self._hf.add_value_vector('lrnz_amp_gen', folder = 'analysis', x = self._frequency_co, unit = 'arb. unit')  
            self._lrnz_f0 = self._hf.add_coordinate('lrnz_f0', folder = 'analysis', unit = 'Hz')
            self._lrnz_k = self._hf.add_coordinate('lrnz_k', folder = 'analysis', unit = 'Hz')
            self._lrnz_a = self._hf.add_coordinate('lrnz_a', folder = 'analysis', unit = '')
            self._lrnz_offs = self._hf.add_coordinate('lrnz_offs', folder = 'analysis', unit = '')
            self._lrnz_Ql = self._hf.add_coordinate('lrnz_ql', folder = 'analysis', unit = '')
    
            self._lrnz_chi2_fit  = self._hf.add_coordinate('lrnz_chi2' , folder = 'analysis', unit = '')
            
        if self._ds_type == ds_types['matrix']: # data from measure_2d
            self._lrnz_amp_gen = self._hf.add_value_matrix('lrnz_amp_gen', folder = 'analysis', x = self._x_co, y = self._frequency_co, unit = 'arb. unit')
            self._lrnz_f0 = self._hf.add_value_vector('lrnz_f0', folder = 'analysis', x = self._x_co, unit = 'Hz')
            self._lrnz_k = self._hf.add_value_vector('lrnz_k', folder = 'analysis', x = self._x_co, unit = 'Hz')
            self._lrnz_a = self._hf.add_value_vector('lrnz_a', folder = 'analysis', x = self._x_co, unit = '')
            self._lrnz_offs = self._hf.add_value_vector('lrnz_offs', folder = 'analysis', x = self._x_co, unit = '')
            self._lrnz_Ql = self._hf.add_value_vector('lrnz_ql', folder = 'analysis', x = self._x_co, unit = '')
    
            self._lrnz_chi2_fit  = self._hf.add_value_vector('lrnz_chi2' , folder = 'analysis', x = self._x_co, unit = '')

        lrnz_view = self._hf.add_view("lrnz_fit", x = self._y_co, y = self._ds_amp)
        lrnz_view.add(x=self._frequency_co, y=self._lrnz_amp_gen)

    def _lorentzian_from_fit(self,fit):
        return fit[2] / (1 + (4*((self._fit_frequency-fit[0])/fit[1]) ** 2)) + fit[3]

    def _lorentzian_fit_chi2(self, fit, amplitudes_sq):
        chi2 = np.sum((self._lorentzian_from_fit(fit)-amplitudes_sq)**2) / (len(amplitudes_sq)-len(fit))
        return chi2

    def fit_skewed_lorentzian(self, fit_all = False, f_min=None, f_max=None,pre_filter_data=None):
        '''
        skewed lorentzian fit for amp data in the f_min-f_max frequency range
        squared amps are fitted at skewed lorentzian using scipy.leastsq
        fit parameter, chi2, and generated amp are stored in the hdf-file

        input:
        fit_all (bool): True or False, default: False. Whole data (True) or only last "slice" (False) is fitted (optional)
        f_min (float): lower boundary for data to be fitted (optional, default: None, results in min(frequency-array))
        f_max (float): upper boundary for data to be fitted (optional, default: None, results in max(frequency-array))
        '''
        def residuals(p,x,y):
            A2, A4, Qr = p
            err = y -(A1a+A2*(x-fra)+(A3a+A4*(x-fra))/(1.+4.*Qr**2*((x-fra)/fra)**2))
            return err
        def residuals2(p,x,y):
            A1, A2, A3, A4, fr, Qr = p
            err = y -(A1+A2*(x-fr)+(A3+A4*(x-fr))/(1.+4.*Qr**2*((x-fr)/fr)**2))
            return err

        self._fit_all = fit_all

        if not self._datasets_loaded:
            self._get_datasets()
        self._update_data()

        self._prepare_f_range(f_min,f_max)
        if self._first_skewed_lorentzian:
            self._prepare_skewed_lorentzian()
            self._first_skewed_lorentzian = False

        '''
        fit_amplitude is always 2dim np array.
        for 1dim data, shape: (1, # fit frequency points)
        for 2dim data, shape: (# amplitude slices (i.e. power values in scan), # fit frequency points)
        '''
        if not self._fit_all:
            self._get_last_amp_trace()

        for amplitudes in self._fit_amplitude:
            "fits a skewed lorenzian to reflection amplitudes of a resonator"
            # prefilter the data
            amplitudes = self._pre_filter_data(amplitudes)

            amplitudes = np.absolute(amplitudes)
            amplitudes_sq = amplitudes**2

            A1a = np.minimum(amplitudes_sq[0],amplitudes_sq[-1])
            A3a = -np.max(amplitudes_sq)
            fra = self._fit_frequency[np.argmin(amplitudes_sq)]

            p0 = [0., 0., 1e3]

            try:
                p_final = leastsq(residuals,p0,args=(self._fit_frequency,amplitudes_sq))
                A2a, A4a, Qra = p_final[0]

                p0 = [A1a, A2a , A3a, A4a, fra, Qra]
                p_final = leastsq(residuals2,p0,args=(self._fit_frequency,amplitudes_sq))
                popt=p_final[0]
            except:
                self._skwd_amp_gen.append(np.array([np.nan for f in self._fit_frequency]))
                self._skwd_f0.append(np.nan)
                self._skwd_a1.append(np.nan)
                self._skwd_a2.append(np.nan)
                self._skwd_a3.append(np.nan)
                self._skwd_a4.append(np.nan)
                self._skwd_Qr.append(np.nan)
                self._skwd_chi2_fit.append(np.nan)
            else:
                chi2 = self._skewed_fit_chi2(popt,amplitudes_sq)
                amp_gen = np.sqrt(np.array(self._skewed_from_fit(popt)))

                self._skwd_amp_gen.append(amp_gen)
                self._skwd_f0.append(float(popt[4]))
                self._skwd_a1.append(float(popt[0]))
                self._skwd_a2.append(float(popt[1]))
                self._skwd_a3.append(float(popt[2]))
                self._skwd_a4.append(float(popt[3]))
                self._skwd_Qr.append(float(popt[5]))
                self._skwd_chi2_fit.append(float(chi2))

            self._skwd_Qi.append(self._skewed_estimate_Qi(popt))

    def _prepare_skewed_lorentzian(self):
        '''
        creates the datasets for the skewed lorentzian fit in the hdf-file
        '''
        if self._ds_type == ds_types['vector']: # data from measure_1d
            self._skwd_amp_gen = self._hf.add_value_vector('sklr_amp_gen', folder = 'analysis', x = self._frequency_co, unit = 'arb. unit')
            self._skwd_f0 = self._hf.add_coordinate('sklr_f0', folder = 'analysis', unit = 'Hz')
            self._skwd_a1 = self._hf.add_coordinate('sklr_a1', folder = 'analysis', unit = 'Hz')
            self._skwd_a2 = self._hf.add_coordinate('sklr_a2', folder = 'analysis', unit = '')
            self._skwd_a3 = self._hf.add_coordinate('sklr_a3', folder = 'analysis', unit = '')
            self._skwd_a4 = self._hf.add_coordinate('sklr_a4', folder = 'analysis', unit = '')
            self._skwd_Qr = self._hf.add_coordinate('sklr_qr', folder = 'analysis', unit = '')
            self._skwd_Qi = self._hf.add_coordinate('sklr_qi', folder = 'analysis', unit = '')
    
            self._skwd_chi2_fit  = self._hf.add_coordinate('sklr_chi2' , folder = 'analysis', unit = '')

        if self._ds_type == ds_types['matrix']: # data from measure_2d
            self._skwd_amp_gen = self._hf.add_value_matrix('sklr_amp_gen', folder = 'analysis', x = self._x_co, y = self._frequency_co, unit = 'arb. unit')
            self._skwd_f0 = self._hf.add_value_vector('sklr_f0', folder = 'analysis', x = self._x_co, unit = 'Hz')
            self._skwd_a1 = self._hf.add_value_vector('sklr_a1', folder = 'analysis', x = self._x_co, unit = 'Hz')
            self._skwd_a2 = self._hf.add_value_vector('sklr_a2', folder = 'analysis', x = self._x_co, unit = '')
            self._skwd_a3 = self._hf.add_value_vector('sklr_a3', folder = 'analysis', x = self._x_co, unit = '')
            self._skwd_a4 = self._hf.add_value_vector('sklr_a4', folder = 'analysis', x = self._x_co, unit = '')
            self._skwd_Qr = self._hf.add_value_vector('sklr_qr', folder = 'analysis', x = self._x_co, unit = '')
            self._skwd_Qi = self._hf.add_value_vector('sklr_qi', folder = 'analysis', x = self._x_co, unit = '')
    
            self._skwd_chi2_fit  = self._hf.add_value_vector('sklr_chi2' , folder = 'analysis', x = self._x_co, unit = '')

        skwd_view = self._hf.add_view('sklr_fit', x = self._y_co, y = self._ds_amp)
        skwd_view.add(x=self._frequency_co, y=self._skwd_amp_gen)

    def _skewed_fit_chi2(self, fit, amplitudes_sq):
        chi2 = np.sum((self._skewed_from_fit(fit)-amplitudes_sq)**2) / (len(amplitudes_sq)-len(fit))
        return chi2

    def _skewed_from_fit(self,p):
        A1, A2, A3, A4, fr, Qr = p
        return A1+A2*(self._fit_frequency-fr)+(A3+A4*(self._fit_frequency-fr))/(1.+4.*Qr**2*((self._fit_frequency-fr)/fr)**2)

    def _skewed_estimate_Qi(self,p):

        #this is a very clumsy numerical estimate of the Qi factor based on the +3dB method.#
        A1, A2, A3, A4, fr, Qr = p
        def skewed_from_fit(p,f):
            A1, A2, A3, A4, fr, Qr = p
            return A1+A2*(f-fr)+(A3+A4*(f-fr))/(1.+4.*Qr**2*((f-fr)/fr)**2)

        #df = fr/(Qr*10000)
        fmax = fr+fr/Qr
        fs = np.linspace(fr,fmax,1000,dtype=np.float64)
        Amin = skewed_from_fit(p,fr)
        #print("---")
        #print("Amin, 2*Amin",Amin, 2*Amin)

        for f in fs:
            A = skewed_from_fit(p,f)
            #print(A, f)
            if A>2*Amin:
                break
        qi = fr/(2*(f-fr))
        #print("f, A, fr/2*(f-fr)", f,A, qi)

        return float(qi)

    def _prepare_fano(self):
        "create the datasets for the fano fit in the hdf-file"
        if self._ds_type == ds_types['vector']: # data from measure_1d
            self._fano_amp_gen = self._hf.add_value_vector('fano_amp_gen', folder = 'analysis', x = self._frequency_co, unit = 'arb. unit')
            self._fano_q_fit  = self._hf.add_coordinate('fano_q' , folder = 'analysis', unit = '')
            self._fano_bw_fit = self._hf.add_coordinate('fano_bw', folder = 'analysis', unit = 'Hz')
            self._fano_fr_fit = self._hf.add_coordinate('fano_fr', folder = 'analysis', unit = 'Hz')
            self._fano_a_fit  = self._hf.add_coordinate('fano_a' , folder = 'analysis', unit = '')
    
            self._fano_chi2_fit  = self._hf.add_coordinate('fano_chi2' , folder = 'analysis', unit = '')
            self._fano_Ql_fit    = self._hf.add_coordinate('fano_Ql' , folder = 'analysis', unit = '')
            self._fano_Q0_fit    = self._hf.add_coordinate('fano_Q0' , folder = 'analysis', unit = '')
            
        if self._ds_type == ds_types['matrix']: # data from measure_2d
            self._fano_amp_gen = self._hf.add_value_matrix('fano_amp_gen', folder = 'analysis', x = self._x_co, y = self._frequency_co, unit = 'arb. unit')
            self._fano_q_fit  = self._hf.add_value_vector('fano_q' , folder = 'analysis', x = self._x_co, unit = '')
            self._fano_bw_fit = self._hf.add_value_vector('fano_bw', folder = 'analysis', x = self._x_co, unit = 'Hz')
            self._fano_fr_fit = self._hf.add_value_vector('fano_fr', folder = 'analysis', x = self._x_co, unit = 'Hz')
            self._fano_a_fit  = self._hf.add_value_vector('fano_a' , folder = 'analysis', x = self._x_co, unit = '')
    
            self._fano_chi2_fit  = self._hf.add_value_vector('fano_chi2' , folder = 'analysis', x = self._x_co, unit = '')
            self._fano_Ql_fit    = self._hf.add_value_vector('fano_Ql' , folder = 'analysis', x = self._x_co, unit = '')
            self._fano_Q0_fit    = self._hf.add_value_vector('fano_Q0' , folder = 'analysis', x = self._x_co, unit = '')

        fano_view = self._hf.add_view('fano_fit', x = self._y_co, y = self._ds_amp)
        fano_view.add(x=self._frequency_co, y=self._fano_amp_gen)

    def fit_fano(self,fit_all = False, f_min=None, f_max=None,pre_filter_data=None):
        '''
        fano fit for amp data in the f_min-f_max frequency range
        squared amps are fitted at fano using scipy.leastsq
        fit parameter, chi2, q0, and generated amp are stored in the hdf-file

        input:
        fit_all (bool): True or False, default: False. Whole data (True) or only last "slice" (False) is fitted (optional)
        f_min (float): lower boundary for data to be fitted (optional, default: None, results in min(frequency-array))
        f_max (float): upper boundary for data to be fitted (optional, default: None, results in max(frequency-array))
        '''

        self._fit_all = fit_all
        if not self._datasets_loaded:
            self._get_datasets()
        self._update_data()
        self._prepare_f_range(f_min,f_max)
        if self._first_fano:
            self._prepare_fano()
            self._first_fano = False

        '''
        fit_amplitude is always 2dim np array.
        for 1dim data, shape: (1, # fit frequency points)
        for 2dim data, shape: (# amplitude slices (i.e. power values in scan), # fit frequency points)
        '''
        if not self._fit_all:
            self._get_last_amp_trace()

        for amplitudes in self._fit_amplitude:
            amplitude_sq = (np.absolute(amplitudes))**2
            try:
                fit = self._do_fit_fano(amplitude_sq)
                amplitudes_gen = self._fano_reflection_from_fit(fit)

                '''calculate the chi2 of fit and data'''
                chi2 = self._fano_fit_chi2(fit, amplitude_sq)

            except:
                self._fano_amp_gen.append(np.array([np.nan for f in self._fit_frequency]))
                self._fano_q_fit.append(np.nan)
                self._fano_bw_fit.append(np.nan)
                self._fano_fr_fit.append(np.nan)
                self._fano_a_fit.append(np.nan)
                self._fano_chi2_fit.append(np.nan)
                self._fano_Ql_fit.append(np.nan)
                self._fano_Q0_fit.append(np.nan)

            else:
                ''' save the fitted data to the hdf_file'''
                self._fano_amp_gen.append(np.sqrt(np.absolute(amplitudes_gen)))
                self._fano_q_fit.append(float(fit[0]))
                self._fano_bw_fit.append(float(fit[1]))
                self._fano_fr_fit.append(float(fit[2]))
                self._fano_a_fit.append(float(fit[3]))
                self._fano_chi2_fit.append(float(chi2))
                self._fano_Ql_fit.append(float(fit[2])/float(fit[1]))
                q0=self._fano_fit_q0(np.sqrt(np.absolute(amplitudes_gen)),float(fit[2]))
                self._fano_Q0_fit.append(q0)

    def _fano_reflection(self,f,q,bw,fr,a=1,b=1):
        '''
        evaluates the fano function in reflection at the
        frequency f
        with
        resonator frequency fr
        attenuation a (linear)
        fano-factor q
        bandwidth bw
        '''
        return a*(1 - self._fano_transmission(f,q,bw,fr))

    def _fano_transmission(self,f,q,bw,fr,a=1,b=1):
        '''
        evaluates the normalized transmission fano function at the
        frequency f
        with
        resonator frequency fr
        attenuation a (linear)
        fano-factor q
        bandwidth bw
        '''
        F = 2*(f-fr)/bw
        return ( 1/(1+q**2) * (F+q)**2 / (F**2+1))

    def _do_fit_fano(self, amplitudes_sq):
        # initial guess
        bw = 1e6
        q  = 1 #np.sqrt(1-amplitudes_sq).min()  # 1-Amp_sq = 1-1+q^2  => A_min = q
        fr = self._fit_frequency[np.argmin(amplitudes_sq)]
        a  = amplitudes_sq.max()

        p0 = [q, bw, fr, a]

        def fano_residuals(p,frequency,amplitude_sq):
            q, bw, fr, a = p
            err = amplitude_sq-self._fano_reflection(frequency,q,bw,fr=fr,a=a)
            return err

        p_fit = leastsq(fano_residuals,p0,args=(self._fit_frequency,np.array(amplitudes_sq)))
        #print(("q:%g bw:%g fr:%g a:%g")% (p_fit[0][0],p_fit[0][1],p_fit[0][2],p_fit[0][3]))
        return p_fit[0]

    def _fano_reflection_from_fit(self,fit):
        return self._fano_reflection(self._fit_frequency,fit[0],fit[1],fit[2],fit[3])

    def _fano_fit_chi2(self,fit,amplitudes_sq):
        chi2 = np.sum((self._fano_reflection_from_fit(fit)-amplitudes_sq)**2) / (len(amplitudes_sq)-len(fit))
        return chi2

    def _fano_fit_q0(self,amp_gen,fr):
        '''
        calculates q0 from 3dB bandwidth above minimum in fit function
        '''
        amp_3dB=10*np.log10((np.min(amp_gen)))+3
        amp_3dB_lin=10**(amp_3dB/10)
        f_3dB=[]
        for i in range(len(amp_gen)-1):
            if np.sign(amp_gen[i]-amp_3dB_lin) != np.sign(amp_gen[i+1]-amp_3dB_lin):#crossing@amp_3dB
                f_3dB.append(self._fit_frequency[i])
        if len(f_3dB)>1:
            q0 = fr/(f_3dB[1]-f_3dB[0])
            return float(q0)
        else: return np.nan

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(
        description="resonator.py hdf-based simple resonator fit frontend / KIT 2015")

    parser.add_argument('-f','--file',     type=str, help='hdf filename to open')
    parser.add_argument('-lf','--lorentzian-fit',  default=False,action='store_true', help='(optional) lorentzian fit')
    parser.add_argument('-ff','--fano-fit',     default=False,action='store_true', help='(optional) fano fit')
    parser.add_argument('-cf','--circle-fit',   default=False,action='store_true', help='(optional) circle fit')
    parser.add_argument('-slf','--skewed-lorentzian-fit', default=False,action='store_true',help='(optional) skewed lorentzian fit')
    parser.add_argument('-all','--fit-all', default=False,action='store_true',help='(optional) fit all entries in dataset')
    parser.add_argument('-fr','--frequency-range', type=str, help='(optional) frequency range for fitting, comma separated')
    parser.add_argument('-fg','--filter-gaussian', default=False, action='store_true', help='(optional) (pre-) filter data: gaussian')
    parser.add_argument('-fm','--filter-median', default=False, action='store_true', help='(optional) (pre-) filter data: median')
    parser.add_argument('-fp','--filter-params',type=str, help='(optional) (pre-) filter data: parameter')
    parser.add_argument('-d','--debug-output', default=False, action='store_true', help='(optional) debug: more verbose')
    parser.add_argument('-t','--type',   type=str, help='resonator type: (r)eflection or (n)otch')
    args=parser.parse_args()
    #argsfile=None
    if args.file:
        R = Resonator(args.file)
        if args.debug_output:
            R._debug = True

        fit_all = args.fit_all

        if args.frequency_range:
            freq_range=args.frequency_range.split(',')
            f_min=int(float(freq_range[0]))
            f_max=int(float(freq_range[1]))
        else:
            f_min=None
            f_max=None

        if args.filter_median:
            if args.filter_params:
                filter_params = args.filter_params.split(',')
                R.set_prefilter(median=True,params=[float(filter_params[0])])
            else:
                R.set_prefilter(median=True)

        if args.filter_gaussian:
            if args.filter_params:
                filter_params = args.filter_params.split(',')
                R.set_prefilter(gaussian=True,params=[float(filter_params[0])])
            else:
                R.set_prefilter(gaussian=True)

        if args.circle_fit:
            if args.type == 'r':
                R.fit_circle(reflection = True, fit_all=fit_all, f_min=f_min,f_max=f_max)
            elif args.type == 'n':
                R.fit_circle(notch = True, fit_all=fit_all, f_min=f_min,f_max=f_max)
            else:
                R.fit_circle(notch = True, fit_all=fit_all, f_min=f_min,f_max=f_max)
        if args.lorentzian_fit:
            R.fit_lorentzian(fit_all=fit_all, f_min=f_min,f_max=f_max)
        if args.skewed_lorentzian_fit:
            R.fit_skewed_lorentzian(fit_all=fit_all, f_min=f_min,f_max=f_max)
        if args.fano_fit:
            R.fit_fano(fit_all=fit_all, f_min=f_min,f_max=f_max)
        R.close()
    else:
        print("no file supplied. type -h for help")



# Quantum machines coherence library
# started by M. Spiecker, 06.2020
#
# Quantum machines config class for single qubit experiments with an interferrometric setup

import numpy as np
from qm.QuantumMachinesManager import QuantumMachinesManager

from qkit.measure.timedomain.quantum_machine import pulses


class QMQubitConfig:

    def __init__(self):

        self.qmm = QuantumMachinesManager()
        self.qm = None

        self.config = None

        # Qubit
        self.qubit_freq = 1e9
        self.QB_IF_freq = 80.0e6
        self.QB_LO_freq = self.qubit_freq - self.QB_IF_freq  # for the mixer

        self.saturation_pulse_len = 50000
        self.pi_pulse_len = 48
        self.pi_wf = pulses.gauss(0.49, 0.0, 10.0, self.pi_pulse_len)
        self.pi2_wf = self.pi_wf / 2

        # Readout
        self.res_freq = 7e9
        self.RR_IF_freq = 62.5e6
        self.RR_LO_up_freq = self.res_freq + self.RR_IF_freq
        self.RR_LO_down_freq = self.res_freq - self.RR_IF_freq

        self.wait_sig = 160  #  minimum 32 clock cycles
        self.smearing_sig = 0  # multiple period
        self.wait_ref = 53  # minimum 32 clock cycles
        self.smearing_ref = 0

        self.readout_pulse_len = 512  # multiple of 16
        self.readout_wf = pulses.const(0.49, self.readout_pulse_len)
        self.theta = 0.0  # rotation of IQ plane

        # create config and load it
        self.load_config()

    def set_readout_wf(self, func, args, plot=False):

        self.readout_wf = func(*args, self.readout_pulse_len)
        self.load_config()

        #if plot:
        #    self.ax.plot(self.readout_wf)
        #    plt.show()

    def set_pi_pulse_wf(self, func, args, plot=False):

        self.pi_wf = func(*args, self.pi_pulse_len)
        self.pi2_wf = self.pi_wf / 2
        self.load_config()

        #if plot:
        #    self.ax.plot(self.pi_wf)
        #    self.ax.plot(self.pi2_wf)
        #    plt.show()

    def set_theta(self, theta):

        self.theta = theta
        self.load_config()


    # TODO set dc offset mixers
    def load_config(self):

        opx_one = "con1"

        self.config = {
            'version': 1,
            'controllers': {
                opx_one: {
                    'type': 'Opx',
                    'analog_outputs': {
                        # sweep offset of I,Q to see which combination best suppresses LO leakage
                        1: {'offset': 0},  # I qubit
                        2: {'offset': 0},  # Q qubit
                        3: {'offset': 0},  # RR
                    },
                    'analog_inputs': {
                        1: {'offset': 0},  # Reference
                        2: {'offset': 0}  # Signal
                    },
                    'digital_outputs': {
                        1: {}
                    }
                },
            },
            'elements': {
                'signal': {
                    "singleInput": {
                        "port": (opx_one, 3),
                    },
                    'intermediate_frequency': self.RR_IF_freq,
                    'operations': {
                        'readout_sig': 'readout_pulse',
                        'saturation_pulse': 'readout_saturation_pulse',
                    },
                    "outputs": {
                        'out1': (opx_one, 1),
                    },
                    'time_of_flight': 4 * self.wait_sig,
                    'smearing': self.smearing_sig
                },
                'reference': {
                    "singleInput": {
                        "port": (opx_one, 3),
                    },
                    'intermediate_frequency': self.RR_IF_freq,
                    'operations': {
                        'readout_ref': 'reference_pulse',
                    },
                    "outputs": {
                        'out1': (opx_one, 2),  # TODO out2?
                    },
                    'time_of_flight': 4 * self.wait_ref,
                    'smearing': self.smearing_ref
                },
                'qubit': {
                    "mixInputs": {
                        "I": (opx_one, 1),
                        "Q": (opx_one, 2),
                        "mixer": "mixer_QB",
                        "lo_frequency": self.QB_LO_freq
                    },
                    'intermediate_frequency': self.QB_IF_freq,
                    'operations': {
                        'saturation_pulse': 'saturation_pulse',
                        'pi_pulse': 'pi_pulse',
                        'pi/2_pulse': 'pi/2_pulse',
                    }
                },
            },

            "pulses": {
                'readout_saturation_pulse': {
                    'operation': 'control',
                    'length': self.saturation_pulse_len,
                    'waveforms': {
                        'single': 'const_wf',
                    },
                },
                'saturation_pulse': {
                    'operation': 'control',
                    'length': self.saturation_pulse_len,
                    'waveforms': {
                        'I': 'const_wf',
                        'Q': 'zero_wf',
                    }
                },
                'pi_pulse': {
                    'operation': 'control',
                    'length': self.pi_pulse_len,
                    'waveforms': {
                        'I': 'pi_wf',
                        'Q': 'zero_wf'
                    },
                },
                'pi/2_pulse': {
                    'operation': 'control',
                    'length': self.pi_pulse_len,
                    'waveforms': {
                        'I': 'pi/2_wf',
                        'Q': 'zero_wf'
                    },
                },
                'readout_pulse': {
                    'operation': 'measurement',
                    'length': self.readout_pulse_len,
                    'waveforms': {
                        'single': 'readout_wf',
                    },
                    'digital_marker': 'ON',
                    'integration_weights': {
                        'integW_Is': 'integW_Is',
                        'integW_Qs': 'integW_Qs'
                    },
                },
                'reference_pulse': {
                    'operation': 'measurement',
                    'length': self.readout_pulse_len,
                    'waveforms': {
                        'single': 'zero_wf',
                    },
                    'digital_marker': 'ON',
                    'integration_weights': {
                        'integW_Ir': 'integW_Ir',
                        'integW_Qr': 'integW_Qr',
                    }
                },
            },

            'waveforms': {
                'zero_wf': {
                    'type': 'constant',
                    'sample': 0
                },
                'const_wf': {
                    'type': 'constant',
                    'sample': 0.49
                },
                'readout_wf': {
                    'type': 'arbitrary',
                    'samples': self.readout_wf.tolist()
                },
                'pi_wf': {
                    'type': 'arbitrary',
                    'samples': self.pi_wf.tolist()
                },
                'pi/2_wf': {
                    'type': 'arbitrary',
                    'samples': self.pi2_wf.tolist()
                }
            },
            'digital_waveforms': {
                'ON': {
                    'samples': [(1, 0)]
                },
            },

            'integration_weights': {
                'integW_Is': {
                    'cosine': [np.cos(self.theta / 360 * np.pi)] * int(self.readout_pulse_len / 4 + self.smearing_sig),
                    'sine': [np.sin(self.theta / 360 * np.pi)] * int(self.readout_pulse_len / 4 + self.smearing_sig),
                },
                'integW_Qs': {
                    'cosine': [np.sin(- self.theta / 360 * np.pi)] * int(
                        self.readout_pulse_len / 4 + self.smearing_sig),
                    'sine': [np.cos(self.theta / 360 * np.pi)] * int(self.readout_pulse_len / 4 + self.smearing_sig),
                },
                'integW_Ir': {
                    'cosine': [np.cos(self.theta / 360 * np.pi)] * int(self.readout_pulse_len / 4 + self.smearing_ref),
                    'sine': [np.sin(- self.theta / 360 * np.pi)] * int(self.readout_pulse_len / 4 + self.smearing_ref),
                },
                'integW_Qr': {
                    'cosine': [np.sin(self.theta / 360 * np.pi)] * int(self.readout_pulse_len / 4 + self.smearing_ref),
                    'sine': [np.cos(self.theta / 360 * np.pi)] * int(self.readout_pulse_len / 4 + self.smearing_ref),
                },
            },
            'mixers': {
                'mixer_QB': [
                    {
                        'intermediate_frequency': self.QB_IF_freq,
                        'lo_frequency': self.QB_LO_freq,
                        'correction': [1, 0, 0, 1]
                    }
                ],
            },
        }

        self.qm = self.qmm.open_qm(self.config)


    # TODO make it nice
    def print_config(self):
        print(self.config)



# M2I2030.py driver class, to perform the communication between the Wrapper and the card
# Pieter de Groot <pieterdegroot@gmail.com>, 2008
# Martijn Schaafsma <qtlab@mcschaafsma.nl>, 2008
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from ctypes import *
from _Spectrum_M3i2132.errors import errors as _spcm_errors
from _Spectrum_M3i2132.regs import regs as _spcm_regs
from qkit.core.instrument_base import Instrument
import pickle
from time import sleep, time
import types
import logging
import numpy
import platform

class Spectrum_M3i2132(Instrument):
    '''
    This is the driver for the Spectrum M3i2132 data acquisition card

    Usage:
    Initialize with
    <name> = instruments.create('name', 'Spectrum_M3i2132')

    TODO:
    1) Fix get all
    2) Fix readout modes
    3) Fix representation and organisation of data
    4) Readout of two channels
    5) Add self._cardopened oid ??
    6) incosistent use of set_
    7) fix handling of timeout! (not enough triggers detected) (error nr 263)
    '''

    def __init__(self, name):
        '''
        Initializes the dataacquisition card, and communicates with the wrapper.

        Usage:
            Use in a simple measurementloop as:
            <name>.init_default(memsize, posttrigger, amp)

            And repeat:
            <name>.start_with_trigger_and_waitready()
            <name>.readout_singlemode_float()

        Input:
            name (string) : name of the instrument

        Output:
            None
        '''
        # Initialize wrapper
        logging.info(__name__ + ' : Initializing instrument Spectrum')
        Instrument.__init__(self, name, tags=['physical'])

        # Load dll and open connection
        self._card_is_open = False
        self._load_dll()
        self._open()

        # add parameters
        self.add_parameter('timeout', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('trigger_delay', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('memsize', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('post_trigger', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('input_offset_ch0', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('input_offset_ch1', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('input_amp_ch0', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('input_amp_ch1', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('input_path_ch0', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('input_path_ch1', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('spc_samplerate', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('reference_clock', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('segmentsize', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('loops', flags=Instrument.FLAG_GETSET, type=int)
        self.add_parameter('trigger_ext0_level0', flags=Instrument.FLAG_SET, type=int, minval=-5000, maxval=5000, unit='mV')
        self.add_parameter('trigger_ext0_level1', flags=Instrument.FLAG_SET, type=int, minval=-5000, maxval=5000, unit='mV')

        self.add_parameter('serial', flags=Instrument.FLAG_GET)
        self.add_parameter('ramsize', flags=Instrument.FLAG_GET)
        self.add_parameter('card_status', flags=Instrument.FLAG_GET)

        # add functions
        self.add_function('start')
        self.add_function('start_with_trigger_and_waitready')
        self.add_function('reset')
        self.add_function('writesetup')
        self.add_function('enable_trigger')
        self.add_function('force_trigger')
        self.add_function('disable_trigger')
        self.add_function('stop')
        self.add_function('waitprefull')
        self.add_function('waittrigger')
        self.add_function('waitready')
        self.add_function('select_channel0')
        self.add_function('select_channel1')
        self.add_function('select_channel01')
        self.add_function('input_term_ch0_50Ohm')
        self.add_function('input_term_ch0_1MOhm')
        self.add_function('input_term_ch1_50Ohm')
        self.add_function('input_term_ch1_1MOhm')
        self.add_function('set_clock_50Ohm')
        self.add_function('set_clock_highOhm')
        self.add_function('set_clockmode_pll')
        self.add_function('set_clockmode_quartz1')
        self.add_function('set_single_mode')
        self.add_function('set_multi_mode')
        self.add_function('trigger_mode_pos')
        self.add_function('trigger_mode_neg')
        self.add_function('set_trigger_ORmask_tmask_ext0')
        self.add_function('set_trigger_ext0_AND_ext1')
        self.add_function('set_trigger_ext0')
        self.add_function('trigger_termination_50Ohm')
        self.add_function('trigger_termination_highOhm')

        self.reset()

    def __del__(self):
        '''
        Closes up the Spectrum driver

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : Deleting Spectrum instrument')
        self._close()

###########################
### init related functions
###########################

    def _load_dll(self):
        '''
        Loads the functions from spcm_win32.dll

        Input:
            None

        Output:
            None
        '''

        if platform.architecture()[0] == '64bit': self.pf_64Bit = True
        else: self.pf_64Bit = False

        if self.pf_64Bit:
            logging.debug(__name__ + ' : Loading spcm_win64.dll')
            drv_handle= POINTER(c_uint64)
            self._spcm_win32 = windll.LoadLibrary('C:\\WINDOWS\\System32\\spcm_win64.dll')

            #function names are not decorated in the 64 bit version
            self._spcm_win32.open           = self._spcm_win32["spcm_hOpen"]
            self._spcm_win32.close          = self._spcm_win32["spcm_vClose"]
            self._spcm_win32.SetParam32     = self._spcm_win32["spcm_dwSetParam_i32"]
            self._spcm_win32.SetParam64m    = self._spcm_win32["spcm_dwSetParam_i64m"]
            self._spcm_win32.SetParam64     = self._spcm_win32["spcm_dwSetParam_i64"]
            self._spcm_win32.GetParam32     = self._spcm_win32["spcm_dwGetParam_i32"]
            self._spcm_win32.GetParam64m    = self._spcm_win32["spcm_dwGetParam_i64m"]
            self._spcm_win32.GetParam64     = self._spcm_win32["spcm_dwGetParam_i64"]
            self._spcm_win32.DefTransfer64m = self._spcm_win32["spcm_dwDefTransfer_i64m"]
            self._spcm_win32.DefTransfer64  = self._spcm_win32["spcm_dwDefTransfer_i64"]
            self._spcm_win32.InValidateBuf  = self._spcm_win32["spcm_dwInvalidateBuf"]
            self._spcm_win32.GetErrorInfo   = self._spcm_win32["spcm_dwGetErrorInfo_i32"]
        else:
            drv_handle = c_void_p
            logging.debug(__name__ + ' : Loading spcm_win32.dll')
            self._spcm_win32 = windll.LoadLibrary('C:\\WINDOWS\\System32\\spcm_win32.dll')

            self._spcm_win32.open           = self._spcm_win32["_spcm_hOpen@4"]
            self._spcm_win32.close          = self._spcm_win32["_spcm_vClose@4"]
            self._spcm_win32.SetParam32     = self._spcm_win32["_spcm_dwSetParam_i32@12"]
            self._spcm_win32.SetParam64m    = self._spcm_win32["_spcm_dwSetParam_i64m@16"]
            self._spcm_win32.SetParam64     = self._spcm_win32["_spcm_dwSetParam_i64@16"]
            self._spcm_win32.GetParam32     = self._spcm_win32["_spcm_dwGetParam_i32@12"]
            self._spcm_win32.GetParam64m    = self._spcm_win32["_spcm_dwGetParam_i64m@16"]
            self._spcm_win32.GetParam64     = self._spcm_win32["_spcm_dwGetParam_i64@12"]
            self._spcm_win32.DefTransfer64m = self._spcm_win32["_spcm_dwDefTransfer_i64m@36"]
            self._spcm_win32.DefTransfer64  = self._spcm_win32["_spcm_dwDefTransfer_i64@36"]
            self._spcm_win32.InValidateBuf  = self._spcm_win32["_spcm_dwInvalidateBuf@8"]
            self._spcm_win32.GetErrorInfo   = self._spcm_win32["_spcm_dwGetErrorInfo_i32@16"]
        
        self._spcm_win32.open.argtype = [c_char_p]
        self._spcm_win32.open.restype = drv_handle
        self._spcm_win32.close.argtype = [drv_handle]
        self._spcm_win32.close.restype = None
        self._spcm_win32.SetParam32.argtype = [drv_handle, c_int32, c_int32]
        self._spcm_win32.SetParam32.restype = c_uint32
        self._spcm_win32.SetParam64m.argtype = [drv_handle, c_int32, c_int32, c_int32]
        self._spcm_win32.SetParam64m.restype = c_uint32
        self._spcm_win32.SetParam64.argtype = [drv_handle, c_int32, c_int64]
        self._spcm_win32.SetParam64.restye = c_uint32
        self._spcm_win32.GetParam32.argtype = [drv_handle, c_int32, POINTER (c_int32)]
        self._spcm_win32.GetParam32.restype = c_uint32
        self._spcm_win32.GetParam64.argtype = [drv_handle, c_int32, POINTER (c_int64)]
        self._spcm_win32.GetParam64.restype = c_uint32
        self._spcm_win32.DefTransfer64.argtype = [drv_handle, c_uint32, c_uint32, c_uint32, c_void_p, c_uint64, c_uint64]
        self._spcm_win32.DefTransfer64.restype = c_uint32
        self._spcm_win32.InValidateBuf.argtype = [drv_handle, c_uint32]
        self._spcm_win32.InValidateBuf.restype = c_uint32
        self._spcm_win32.GetErrorInfo.argtype = [drv_handle, POINTER (c_uint32), POINTER (c_int32), c_char_p]
        self._spcm_win32.GetErrorInfo.restype = c_uint32
        
    def _open(self):
        '''
        Opens the card, and creates a handle.
        Only execute once.

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Try to open card')
        if ( not self._card_is_open):
          self._spcm_win32.handel = self._spcm_win32.open(create_string_buffer (b'spcm0'))#'spcm0')
          self._card_is_open = True
        else:
          logging.warning(__name__ + ' : Card is already open !')

        if (self._spcm_win32.handel==0):
            logging.error(__name__ + ' : Unable to open card')
            self._card_is_open = False

    def _close(self):
        '''
        Closes the card

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Try to close card')
        self._spcm_win32.close(self._spcm_win32.handel)

############################
### Base communication tools
############################

    def _set_param(self, regnum, regval):
        '''
        Sets the register to the specified value
        Returns the '0', if succesfull,
        returns the errormessage if an error occurred

        Register flags are listed in M2I2030_regs.py

        Input:
            regnum (register flag)       : Flag corresponding with the register
            regval (register flag)       : Flag corresponding with the value to be set


        Output:
            '0' or errormessage (string)
        '''
        logging.debug(__name__ + ' : Set reg %s to %s' %(regnum, regval))
        err = self._spcm_win32.SetParam32(self._spcm_win32.handel, regnum, regval)
        if (err==0):
            return 0
        elif (err == 263):
            logging.error(__name__ + ' : Timeout')
            return 263
        else:
            logging.error(__name__ + ' : Error %s while setting reg %s to %s' % (err, regnum, regval))
            self._get_error()
            raise ValueError('Error communicating with device')

    def _get_param(self, regnum):
        '''
        Reads out a register on the card.
        Returns the register value, if succesfull,
        returns the errormessage if an error occurred

        Register flags are listed in M2I2030_regs.py

        Input:
            regnum (register flag)       : Flag corresponding with the register

        Output:
            value (int)     : Register value
            or
            error (string)  : Error message
        '''
        logging.debug(__name__ + ' : Reading Reg %s' %(regnum))

        val = c_int()
        p_antw = pointer(val)

        err = self._spcm_win32.GetParam32(self._spcm_win32.handel, regnum, p_antw)
        if (err==0):
            return p_antw.contents.value
        else:
            logging.error(__name__ + ' : Error %s while getting reg %s' %(err,regnum))
            self._get_error()
            raise ValueError('Error communicating with device')

    def invalidate_buffer(self, buffertype=_spcm_regs.SPCM_BUF_DATA):
        '''
        Discards the buffer.

        Input:
            buffertype (register flag) : the flag containing the type of buffer.
                                         probably SPCM_BUF_DATA

        Output:
            None
        '''
        logging.debug(__name__ + ' : Invalidating buffer')
        # stop any running DMA transfers
        err = self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_M2CMD,
            _spcm_regs.M2CMD_DATA_STOPDMA)
        if (err!=0):
            logging.error(__name__ + ' : Error %s while setting reg %s to %s' % (err, regnum, regval))
            self._get_error()
            raise ValueError('Error communicating with device')
        # invalidate buffer
        err = self._spcm_win32.InValidateBuf(self._spcm_win32.handel, buffertype)
        if (err==0):
            return 0
        else:
            logging.error(__name__ + ' : Error %s while setting reg %s to %s' % (err, regnum, regval))
            self._get_error()
            raise ValueError('Error communicating with device')

        del self._pbuffer

    def _get_error(self):
        '''
        Reads out and returns the error buffer

        Input:
            None

        Output:
            Errormessage (string)
        '''
        # try to read out error
        logging.debug(__name__ + ' : Reading error')
        j = (c_char * 200)()
        e1 = c_int()
        e2 = c_int()
        p_errortekst = pointer(j)
        p_er1 = pointer(e1)
        p_er2 = pointer(e2)

        self._spcm_win32.GetErrorInfo(self._spcm_win32.handel, p_er1, p_er2, p_errortekst)

        tekst = ""

        for ii in range(200):
            tekst  = tekst + p_errortekst.contents[ii].decode()
        logging.error(__name__ + ' : ' + tekst)
        return tekst



########################
### Card information
########################

    def do_get_serial(self):
        '''
        Reads out the serial number of the card.

        Input:
            None

        Output:
            serial (int) : The serial number
        '''
        logging.debug(__name__ + ' : Reading serial number')
        return self._get_param(_spcm_regs.SPC_PCISERIALNO)

    def do_get_ramsize(self):
        '''
        Returns the size of the random access memory installed
        on the card

        Input:
            None

        Output:
            ramsize (int) : number of bytes
        '''
        logging.debug(__name__ + ' : Reading Ram size')
        return self._get_param(_spcm_regs.SPC_PCIMEMSIZE)


########################################
### self defined set of initial settings
########################################

    def init_channel0_single_mode(self, memsize=2048, posttrigger=1024, amp=500, offs=0):
        '''
        Initiates the card in default single shot readout mode.
        Trigger is set on ext0 with a positive slope
        The buffersize and range are set

        Input:
            memsize (int)   : number of datapoints that are read out
                                default = 2048
            posttrigger(int): number of datapoints taken after the trigger
                                default = 1024
            amp (int)       : half of the range in millivolts
                                default = 500

        Output:
            None
        '''
        logging.debug(__name__ + ' : Initialing card for default single shot readout')

        # Set the trigger modes
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_EXT0_MODE, _spcm_regs.SPC_TM_POS);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_EXT0_PULSEWIDTH, 0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_OUTPUT, 0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_TERM, 0);

        # Set channel information
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_CHENABLE, _spcm_regs.CHANNEL0)
        #self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_CARDMODE, _spcm_regs.SPC_REC_STD_SINGLE)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_CARDMODE, _spcm_regs.SPC_REC_FIFO_SINGLE)
        #self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_LOOPS, 1)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_MEMSIZE, memsize)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_POSTTRIGGER, posttrigger)
        self._spcm_win32.SetParam32(self._spcm_win32_handel, _spcm_regs.SPC_PATH0, 0) # zofo
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_AMP0, amp)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_OFFS0, offs)

        # Set the trigger masks
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_ORMASK, _spcm_regs.SPC_TMASK_EXT0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_ANDMASK,        0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_CH_ORMASK0,     0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_CH_ORMASK1,     0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_CH_ANDMASK0,    0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_CH_ANDMASK1,    0);

        ## Alternative:

        ## Set the trigger modes
        # self.trigger_mode_pos()
        # self.set_trigger_ext0_pulsewidth(0)
        # self.disable_trigger_output()
        ## self.trigger_termination_50Ohm()
        # self.trigger_termination_hihgOhm()


        ## Set channel information
        # self.select_channel0()
        # self.set_single_mode()
        # self.set_memsize(memsize)
        # self.set_post_trigger(posttrigger)
        # self.set_input_amp_ch0(amp)
        # self.set_input_offs_ch0(offs)


        ## Set the trigger masks
        # self.set_trigger_ORmask_tmask_ext0()
        # self.set_trigger_ANDmask_tmask_NO_ext0()
        # self.set_trigger_ORmask_tmask_NO_ch0()
        # self.set_trigger_ORmask_tmask_NO_ch1()
        # self.set_trigger_ANDmask_tmask_NO_ch0()
        # self.set_trigger_ANDmask_tmask_NO_ch1()


    def init_channel0_multiple_recording(self, nums = 1024, segsize=128, posttrigger=1024, amp=500, offs=0):
        '''
        Initiates the card in:
            Standard Multiple Recording mode
            Using only channel 0
            Trigger on ext0, positive slope, 50 Ohm

        Sample rate is left unchanged (default for using two channels is 200 MHz).

        Input:
            nums (int)      : number of consequtive measurements
                                default = 128
            segsize (int)   : number of datapoints that are read out in one shot
                                default = 2048
            posttrigger(int): number of datapoints taken after the trigger
                                default = 1024
            amp (int)       : half of the range in millivolts
                                default = 500

        Output:
            None
        '''
        logging.debug(__name__ + ' : Initialing card for default multiple shot readout')

        memsize = nums*segsize

        # Set the modes
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_EXT0_MODE, _spcm_regs.SPC_TM_POS);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_EXT0_PULSEWIDTH, 0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_OUTPUT, 0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_TERM, 0);

        # Set channel information
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_CHENABLE, _spcm_regs.CHANNEL0)
        #self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_CARDMODE, _spcm_regs.SPC_REC_STD_MULTI)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_CARDMODE, _spcm_regs.SPC_REC_FIFO_MULTI)
        #self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_LOOPS, 1)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_SEGMENTSIZE, segsize)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_MEMSIZE, memsize)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_POSTTRIGGER, posttrigger)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_AMP0, amp)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_OFFS0, offs)

        # Set the masks
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_ORMASK, _spcm_regs.SPC_TMASK_EXT0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_ANDMASK,        0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_CH_ORMASK0,     0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_CH_ORMASK1,     0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_CH_ANDMASK0,    0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_CH_ANDMASK1,    0);

    def init_channel01_multiple_recording(self, rate=100e6, nums=1024, segsize=128, posttrigger=64, amp0=500, offs0=0, amp1=500, offs1=1):
        '''
        Initiates the card in:
            Standard Multiple Recording mode
            Using Channel 0 and 1
            Trigger on ext0, positive slope, 50 Ohm

        Sample rate is left unchanged (default for using two channels is 100 MHz).

        Input:
            nums (int)      : number of consequtive measurements
                                default = 128
            segsize (int)   : number of datapoints that are read out in one shot
                                default = 2048
            posttrigger(int): number of datapoints taken after the trigger
                                default = 1024
            amp (int)       : half of the range in millivolts
                                default = 500

        Output:
            None
        '''
        logging.debug(__name__ + ' : Initialing card for default multiple shot readout')

        memsize = nums*segsize # Note: memsize is defined per channel

        self.set_spc_samplerate(rate)
        self.set_timeout(5000)

        # Set the modes
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_EXT0_MODE, _spcm_regs.SPC_TM_POS);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_EXT0_PULSEWIDTH, 0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_OUTPUT, 0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_TERM, 0);

        # Set channel information
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_CHENABLE, _spcm_regs.CHANNEL0 | _spcm_regs.CHANNEL1)
        #self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_CARDMODE, _spcm_regs.SPC_REC_STD_MULTI)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_CARDMODE, _spcm_regs.SPC_REC_FIFO_MULTI)
        #self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_LOOPS, 1)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_SEGMENTSIZE, segsize)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_MEMSIZE, memsize)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_POSTTRIGGER, posttrigger)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_AMP0, amp0)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_AMP1, amp1)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_OFFS0, offs0)
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_OFFS1, offs1)

        # Set the masks
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_ORMASK, _spcm_regs.SPC_TMASK_EXT0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_ANDMASK,        0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_CH_ORMASK0,     0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_CH_ORMASK1,     0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_CH_ANDMASK0,    0);
        self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_TRIG_CH_ANDMASK1,    0);

#########################
### General
#########################


    def get_all(self):
        logging.debug(__name__ + ' : getting all values from card')

        self.get_input_amp_ch0()
        self.get_input_amp_ch1()
        self.get_input_offset_ch0()
        self.get_input_offset_ch1()

        self.get_memsize()
        self.get_segmentsize()
        self.get_loops()
        self.get_post_trigger()
        self.get_spc_samplerate()
        self.get_reference_clock()
        self.get_trigger_delay()

        self.get_timeout()

        self.get_ramsize()
        self.get_serial()


################################
### Standard card setup commands
################################

    def reset(self):
        '''
        Resets the card to default values

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Reset card')
        self._set_param(_spcm_regs.SPC_M2CMD, _spcm_regs.M2CMD_CARD_RESET)
        self.get_all()

    def writesetup(self):
        '''
        Writes the current setup to the card without starting the hardware.
        This command may be useful if changing some internal settings
        like clock frequency and enabling outputs.

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Write setup enabled')
        self._set_param(_spcm_regs.SPC_M2CMD, _spcm_regs.M2CMD_CARD_WRITESETUP)

    def start(self):
        '''
        Starts the card

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Card started')
        self._buffer_setup()
        self._set_param(_spcm_regs.SPC_M2CMD, _spcm_regs.M2CMD_CARD_START)

    def start_with_trigger(self):
        '''
        Start the card, enables trigger, and waits till
        the trigger went off

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Card started with trigger')
        self._buffer_setup()
        self._set_param(_spcm_regs.SPC_M2CMD,
            _spcm_regs.M2CMD_CARD_START | _spcm_regs.M2CMD_CARD_ENABLETRIGGER | _spcm_regs.M2CMD_DATA_STARTDMA)

    def start_with_trigger_and_waitready(self):
        '''
        Start the card, enables trigger, and waits till
        the trigger went off

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Card started with trigger and waitready')
        self._buffer_setup()
        self._set_param(_spcm_regs.SPC_M2CMD,
            _spcm_regs.M2CMD_CARD_START | _spcm_regs.M2CMD_CARD_ENABLETRIGGER | _spcm_regs.M2CMD_CARD_WAITREADY | _spcm_regs.M2CMD_DATA_WAITDMA)

    def force_trigger(self):
        '''
        Force a trigger

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Force trigger')
        self._set_param(_spcm_regs.SPC_M2CMD, _spcm_regs.M2CMD_CARD_FORCETRIGGER)

    def stop(self):
        '''
        Stop the card

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Stop card')
        self._set_param(_spcm_regs.SPC_M2CMD, _spcm_regs.M2CMD_CARD_STOP)

    def waitprefull(self):
        '''
        Acquisition modes only: the command waits until the pretrigger
        area has once been filled with data. After pretrigger area
        has been filled the internal trigger engine starts to look for trigger
        events if the trigger detection has been enabled.

        Input:
            None

        Output:
            Error number (0 is no error)
        '''
        logging.debug(__name__ + ' : Wait prefull enabled')
        err = self._set_param(_spcm_regs.SPC_M2CMD, _spcm_regs.M2CMD_WAITPREFULL)
        return err

    def waittrigger(self):
        '''
        Waits until the first trigger event has been detected by the card.
        If using a mode with multiple trigger events like Multiple Recording
        or Gated Sampling there only the first trigger detection will
        generate an interrupt for this wait command.

        Input:
            None

        Output:
            Error number (0 is no error)
        '''
        logging.debug(__name__ + ' : Wait trigger enabled')
        err = self._set_param(_spcm_regs.SPC_M2CMD, _spcm_regs.M2CMD_WAITTRIGGER)
        return err


    def waitready(self):
        '''
        Waits till trigger signal is received

        Input:
            None

        Output:
            Error number (0 is no error)
        '''
        logging.debug(__name__ + ' : Waitready activated')
        err = self._set_param(_spcm_regs.SPC_M2CMD, _spcm_regs.M2CMD_CARD_WAITREADY |  _spcm_regs.M2CMD_DATA_WAITDMA)
        return err


    def do_get_card_status(self):
        '''
        Returns the card status, see p136 of manual

        Input:
            None

        Output:
            status (int): Integer corresponding to the card thatus
        '''
        logging.debug(__name__ + ' : Get card status')
        return self._get_param(_spcm_regs.SPC_M2STATUS)


#################
### channel setup
#################

### select channel

    def select_channel0(self):
        '''
        Select channel 0 for measurement

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Select channel 0')
        self._set_param(_spcm_regs.SPC_CHENABLE, _spcm_regs.CHANNEL0)
        self._numchannels = 1

    def select_channel1(self):
        '''
        Select channel 1 for measurement

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Select channel 1')
        self._set_param(_spcm_regs.SPC_CHENABLE, _spcm_regs.CHANNEL1)
        self._numchannels = 1

    def select_channel01(self):
        '''
        Select channels 0 and 1 for measurement

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Select channels 0 and 1')
        self._set_param(_spcm_regs.SPC_CHENABLE, _spcm_regs.CHANNEL0 | _spcm_regs.CHANNEL1)
        self._numchannels = 2


### set channel termination

    def input_term_ch0_50Ohm(self):
        '''
        Sets the input termination of
        channel 0 to 50 Ohm

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set input termination ch0 to 50 Ohm')
        self.do_set_input_path_ch0(1)

    def input_term_ch0_1MOhm(self):
        '''
        Sets the input termination of
        channel 0 tot 1 MOhm

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set input termination ch0 to 1 MOhm')
        self.do_set_input_path_ch0(0)

    def input_term_ch1_50Ohm(self):
        '''
        Sets the input termination of
        channel 1 to 50 Ohm

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set input termination ch1 to 50 Ohm')
        self.do_set_input_path_ch1(1)

    def input_term_ch1_1MOhm(self):
        '''
        Sets the input termination of
        channel 1 to 1 MOhm

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set input termination ch1 to 1 MOhm')
        self.do_set_input_path_ch1(0)


##################
### Clock settings
##################

### set clock mode

    def set_clockmode_pll(self):
        '''
        Sets the clock mode to PLL

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set clock mode to pll')
        self._set_param(_spcm_regs.SPC_CLOCKMODE, _spcm_regs.SPC_CM_INTPLL)

    def set_clockmode_quartz1(self):
        '''
        Sets the clock mode to quartz1

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set clock mode to quartz1')
        self._set_param(_spcm_regs.SPC_CLOCKMODE, _spcm_regs.SPC_CM_QUARTZ1)

### set clock termination

    def set_clock_50Ohm(self):
        '''
        Sets the clock input termination to 50 Ohm

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set clock termination to 50 Ohm')
        self._set_param(_spcm_regs.SPC_CLOCK50OHM, 1)

    def set_clock_highOhm(self):
        '''
        Sets the clock input termination
        to high impedance

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set clock termination to high impedance')
        self._set_param(_spcm_regs.SPC_CLOCK50OHM, 0)


###########################
### Readout Mode for card
###########################

    def set_single_mode(self):
        '''
        Sets the card in single mode readout status

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set the card in single mode readout status')
        self._set_param(_spcm_regs.SPC_CARDMODE, _spcm_regs.SPC_REC_FIFO_SINGLE)
        #self._set_param(_spcm_regs.SPC_CARDMODE, _spcm_regs.SPC_REC_STD_SINGLE)

    def set_multi_mode(self):
        '''
        Sets the card in 'multiple recording' mode readout status

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set the card in multi mode readout status')
        self._set_param(_spcm_regs.SPC_CARDMODE, _spcm_regs.SPC_REC_FIFO_MULTI)
        #self._set_param(_spcm_regs.SPC_CARDMODE, _spcm_regs.SPC_REC_STD_MULTI)


##############
### Trigger
##############

    def enable_trigger(self):
        '''
        Enables the trigger

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Enable trigger')
        self._set_param(_spcm_regs.SPC_M2CMD, _spcm_regs.M2CMD_CARD_ENABLETRIGGER)

    def disable_trigger(self):
        '''
        Disables the trigger

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Disable trigger')
        self._set_param(_spcm_regs.SPC_M2CMD, _spcm_regs.M2CMD_CARD_DISABLETRIGGER)

    def disable_trigger_output(self):
        '''
        Disables the trigger

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Disable trigger output')
        self._set_param(_spcm_regs.SPC_TRIG_OUTPUT, 0)

### set trigger mode

    def trigger_mode_pos(self):
        '''
        Sets the trigger mode of ext0
        to positive slope

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set trigger mode pos')
        self._set_param(_spcm_regs.SPC_TRIG_EXT0_MODE, _spcm_regs.SPC_TM_POS)

    def trigger_mode_neg(self):
        '''
        Sets the trigger mode of ext0
        to negative slope

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set trigger mode neg')
        self._set_param(_spcm_regs.SPC_TRIG_EXT0_MODE, _spcm_regs.SPC_TM_NEG)

### set trigger properties

    def set_trigger_ext0_pulsewidth(self, width):
        '''
        Sets the pulsewidth for external trigger in samples

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set trigger pulsewidth to %i' % width)
        self._set_param(_spcm_regs.SPC_TRIG_EXT0_PULSEWIDTH, width)

### set trigger mask

    ## ORmask

    def set_trigger_ORmask_tmask_ext0(self):
        '''
        Set trigger OR mask tmask ext0

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set trigger OR mask tmask ext0')
        self._set_param(_spcm_regs.SPC_TRIG_ORMASK, _spcm_regs.SPC_TMASK_EXT0)

    def set_trigger_ORmask_tmask_NO_ch0(self):
        '''
        Set trigger OR mask tmask No ch 0

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set trigger OR mask tmask No ch 0')
        self._set_param(_spcm_regs.SPC_TRIG_CH_ORMASK0, 0)

    def set_trigger_ORmask_tmask_NO_ch1(self):
        '''
        Set trigger OR mask tmask No ch 1

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set trigger OR mask tmask No ch 1')
        self._set_param(_spcm_regs.SPC_TRIG_CH_ORMASK1, 0)

    ## ANDmask

    def set_trigger_ANDmask_tmask_ext0(self):
        '''
        Set trigger AND mask tmask No ext0

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set trigger No mask tmask No ext0')
        self._set_param(_spcm_regs.SPC_TRIG_ANDMASK, 0)

    def set_trigger_ANDmask_tmask_NO_ch0(self):
        '''
        Set trigger AND mask tmask No ch 0

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set trigger AND mask tmask No ch 0')
        self._set_param(_spcm_regs.SPC_TRIG_CH_ANDMASK0, 0)

    def set_trigger_ANDmask_tmask_NO_ch1(self):
        '''
        Set trigger AND mask tmask No ch 1

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set trigger AND mask tmask No ch 1')
        self._set_param(_spcm_regs.SPC_TRIG_CH_ANDMASK1, 0)

### and trigger with ext. sources
    def set_trigger_ext0_AND_ext1(self):
        self._set_param(_spcm_regs.SPCM_X0_MODE, _spcm_regs.SPCM_XMODE_TRIGIN)
        self._set_param(_spcm_regs.SPC_TRIG_EXT1_MODE, _spcm_regs.SPC_TM_HIGH)
        self._set_param(_spcm_regs.SPC_TRIG_ANDMASK, _spcm_regs.SPC_TMASK_EXT0 | _spcm_regs.SPC_TMASK_EXT1)
        self._set_param(_spcm_regs.SPC_TRIG_ORMASK, _spcm_regs.SPC_TMASK_EXT0 | _spcm_regs.SPC_TMASK_EXT1)

    def set_trigger_ext0(self):
        self._set_param(_spcm_regs.SPC_TRIG_ANDMASK, _spcm_regs.SPC_TMASK_EXT0)
        self._set_param(_spcm_regs.SPC_TRIG_ORMASK, _spcm_regs.SPC_TMASK_EXT0)

### trigger levels
    def do_set_trigger_ext0_level0(self, value):
        self._set_param(_spcm_regs.SPC_TRIG_EXT0_LEVEL0, value)

    def do_set_trigger_ext0_level1(self, value):
        self._set_param(_spcm_regs.SPC_TRIG_EXT0_LEVEL1, value)

### set trigger termination

    def trigger_termination_50Ohm(self):
        '''
        Sets the trigger input termination
        to 50 Ohm

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set trigger termination to 50 Ohm')
        self._set_param(_spcm_regs.SPC_TRIG_TERM, 1)

    def trigger_termination_highOhm(self):
        '''
        Sets the trigger input termination
        to high impedance

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set trigger termination to high impedance')
        self._set_param(_spcm_regs.SPC_TRIG_TERM, 0)


#######################
### read data from card
#######################

    def _buffer_setup(self):
        '''
        create a new data buffer
        (assuming the old one is now owned by another part of the program)
        '''
        logging.debug(__name__ + ' : _buffer_setup')
        lMemsize = self.get_memsize()
        lBufsize = lMemsize * self._numchannels

        # setup buffer
        #if hasattr(self, '_pbuffer') and (len(self._pbuffer.contents) == lBufsize):
        #   pass
            # tell the card that the buffer is now available again
            #err = self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_DATA_AVAIL_CARD_LEN,
            #   lBufsize)
            #if (err!=0):
            #   self._get_error()
            #   raise ValueError('Error communicating with device')
        #else:
        a = (c_int8 * lBufsize)()
        p_data = pointer(a)

        # tell card to use buffer
        err = self._spcm_win32.DefTransfer64(self._spcm_win32.handel, _spcm_regs.SPCM_BUF_DATA, 1,
            0, p_data, c_int64(0), c_int64(lBufsize))
        if (err!=0):
            logging.error(__name__ + ' : Error setting up buffer')
            self._get_error()
            raise ValueError('Error communicating with device')

        # start DMA transfers
        err = self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_M2CMD,
            _spcm_regs.M2CMD_DATA_STARTDMA)
        if (err!=0):
            logging.error(__name__ + ' : Error starting DMA transfer, error nr: %i' % err)
            self._get_error()
            raise ValueError('Error communicating with device')

        # save new buffer possibly freeing old one, will break if DMA is in progress
        self._pbuffer = p_data


    def readout_raw_buffer(self, nr_of_channels=1):
        '''
        Reads out the buffer, and returns a list with the size of the
        buffer. Contains only data if the channel is triggered.

        Input:
            None

        Output:
            data (int[memsize]): The data of the buffer
        '''
        logging.debug(__name__ + ' : Readout raw buffer')

        # wait for end of data transfer
        err = self._spcm_win32.SetParam32(self._spcm_win32.handel, _spcm_regs.SPC_M2CMD,
            _spcm_regs.M2CMD_DATA_WAITDMA)# | _spcm_regs.M2CMD_CARD_WAITREADY)
        if (err!=0):
            logging.error(__name__ + ' : Error during read, error nr: %i' % err)
            self._get_error()
            raise ValueError('Error communicating with device')

        return self._pbuffer.contents # [:]

    def readout_singlechannel_singlemode_bin(self):
        '''
        Reads out the buffer, and returns a list with the size of the
        buffer. Contains only data if the channel is triggered.

        Input:
            None

        Output:
            data (int[memsize]): The data of the buffer
        '''
        logging.debug(__name__ + ' : Readout binaries from buffer')

        data = self.readout_raw_buffer()
        return data

    def readout_singlechannel_singlemode_float(self):
        '''
        Reads out the buffer, and converts the data to the actual input voltage.
        Returns a list with the size of the buffer.
        Contains only data if the channel is triggered.

        Input:
            None

        Output:
            dataout (float[memsize]): The data of the buffer
        '''
        logging.debug(__name__ + ' : Readout float after converting from binaries')

        amp = float(self.get_input_amp_ch0())
        offset = float(self.get_input_offset_ch0())

        data = self.readout_raw_buffer()
        if data == 'timeout':
            return data
        data = numpy.array(data, numpy.float32)
        data = []
        data = 2.0 * amp * (data / 255.0) + offset
        return data

    def readout_singlechannel_multimode_bin(self):
        lMemsize = self.get_memsize()
        lSegsize = self.get_segmentsize()

        lnumber_of_segments = int(lMemsize / lSegsize)

        data = self.readout_raw_buffer()
        if data == 'timeout':
            return data
        data = numpy.frombuffer(data, numpy.int8)
        data = numpy.reshape(data, (1, lnumber_of_segments, lSegsize))
        return data

    def readout_singlechannel_multimode_float(self):
        lMemsize = self.get_memsize()
        lSegsize = self.get_segmentsize()
        amp = float(self.get_input_amp_ch0())
        offset = float(self.get_input_offset_ch0())

        lnumber_of_segments = int(lMemsize / lSegsize)

        data = self.readout_raw_buffer()
        if data == 'timeout':
            return data
        data = numpy.array(data, numpy.float32)
        data = numpy.reshape(data, (lnumber_of_segments, lSegsize))
        data = 2.0 * amp * (data / 255.0) + offset
        return data

    def readout_doublechannel_singlemode_bin(self):
        '''
        Reads out the buffer, and returns a list with the size of the
        buffer. Contains only data if the channel is triggered.

        Input:
            None

        Output:
            data (int[memsize]): The data of the buffer
        '''
        logging.debug(__name__ + ' : Readout binaries from buffer')

        lMemsize = self.get_memsize()

        data = self.readout_raw_buffer(nr_of_channels=2)
        if data == 'timeout':
            return data

        data = numpy.array(data, numpy.int8)
        data = numpy.reshape(data, (lMemsize, 2))
        return data

    def readout_doublechannel_multimode_bin(self):
        lMemsize = self.get_memsize()
        lSegsize = self.get_segmentsize()

        lnumber_of_segments = int(lMemsize / lSegsize)

        data = self.readout_raw_buffer(nr_of_channels=2)
        if data == 'timeout':
            return data
        data = numpy.frombuffer(data, numpy.int8, 2*lMemsize)
        data = numpy.reshape(data, (lnumber_of_segments, lSegsize, 2))#(lMemsize, 2))
        data = numpy.rollaxis(data, 2) # channel, segment, sample
        return data

    def readout_doublechannel_multimode_float(self):
        lMemsize = self.get_memsize()
        lSegsize = self.get_segmentsize()
        amp0 = float(self.get_input_amp_ch0())
        offset0 = float(self.get_input_offset_ch1())
        amp1 = float(self.get_input_amp_ch0())
        offset1 = float(self.get_input_offset_ch1())

        lnumber_of_segments = int(lMemsize / lSegsize)

        data = self.readout_raw_buffer(nr_of_channels=2)
        if data == 'timeout':
            return data
        data = numpy.array(data, numpy.int8)
        data = numpy.reshape(data, (lMemsize, 2))
        data0 = data[:,0]
        data1 = data[:,1]
        data0 = numpy.reshape(data0, (lnumber_of_segments, lSegsize))
        data1 = numpy.reshape(data1, (lnumber_of_segments, lSegsize))
        data0 = 2.0 * amp0 * (data0 / 255.0) + offset0
        data1 = 2.0 * amp1 * (data1 / 255.0) + offset1
        return (data0, data1)


### test run

    def test(self, memsize=2048, posttrigger=1024, amp=500):
        '''
        Reads out the buffer, and returns a list with the size of the
        buffer. Contains only data if the channel is triggered.

        Input:
            memsize (int)       : number of datapoints taken
                                    default = 2048
            posttrigger (int)   : numbers of points taken after the trigger
                                    default = 1024
            amp (int)           : half of the range in millivolts
                                    default = 500

        Output:
            data (int[memsize]): Measurement data
        '''
        self.init_default(memsize=memsize, posttrigger=posttrigger, amp=amp)
        print('starting card and waiting for trigger')
        self.start_with_trigger_and_waitready()
        print("received trigger")
        self.data = self.readout_singlemode_float()
        return self.data


###################################
### setting / getting of parameters
###################################

    def do_set_timeout(self, timeout):
        '''
        Set card timeout

        Input:
            timeout (int) : timeout in milliseconds

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set timeout to %s' % timeout)
        self._set_param(_spcm_regs.SPC_TIMEOUT, timeout)

    def do_get_timeout(self):
        '''
        Get card timeout

        Input:
            None

        Output:
            timeout (int) : timeout in milliseconds
        '''
        logging.debug(__name__ + ' : Get timeout')
        return self._get_param(_spcm_regs.SPC_TIMEOUT)

### timing

    def do_set_trigger_delay(self, nums):
        '''
        Set the trigger delay

        Input:
            nums (int) : number of sample clocks delay, must be a multiple of 16

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set trigger delay to %s' % nums)
        self._set_param(_spcm_regs.SPC_TRIG_DELAY, nums)

    def do_get_trigger_delay(self):
        '''
        Get the trigger delay

        Input:
            None

        Output:
            nums (int) : number of sample clocks delay

        '''
        logging.debug(__name__ + ' : Get trigger delay')
        return self._get_param(_spcm_regs.SPC_TRIG_DELAY)

    def do_set_segmentsize(self, lSegsize):
        '''
        Sets the size of the datapoints taken per trigger

        Input:
            lMemsize (int) : number of datapoints, 32+16*n

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set segment size to %s' % lSegsize)
        self._set_param(_spcm_regs.SPC_SEGMENTSIZE, lSegsize)

    def do_get_segmentsize(self):
        '''
        Get the number of datapoints that are read out
        per trigger

        Input:
            None

        Output:
            segmentsize (int) : number of datapoints
        '''
        logging.debug(__name__ + ' : Get segment size')
        return self._get_param(_spcm_regs.SPC_SEGMENTSIZE)

    def do_set_loops(self, lLoops):
        '''
        Sets the number of times the ring buffer is filled.
        #Values other than 1 are effectively ignored.

        Input:
            lMemsize (int) : number of datapoints, 32+16*n

        Output:
            None
        '''
        #lLoops = 1
        logging.debug(__name__ + ' : Set loop count to %s' % lLoops)
        self._set_param(_spcm_regs.SPC_LOOPS, lLoops)

    def do_get_loops(self):
        '''
        Get number of times application ring buffer is filled.

        Input:
            None

        Output:
            segmentsize (int) : number of datapoints
        '''
        logging.debug(__name__ + ' : Get loop count')
        return self._get_param(_spcm_regs.SPC_LOOPS)

    def do_set_post_trigger(self, posttrigger):
        '''
        Sets the number of points that are read out
        after the trigger event

        Input:
            posttrigger (int) : number of points

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set post trigger to %s' % posttrigger)
        self._set_param( _spcm_regs.SPC_POSTTRIGGER, posttrigger)

    def do_get_post_trigger(self):
        '''
        Gets the number of points that are read out
        after the trigger event

        Input:
            None

        Output:
            posttrigger (int) : number of points
        '''
        logging.debug(__name__ + ' : Get post trigger')
        return self._get_param( _spcm_regs.SPC_POSTTRIGGER)

### buffer

    def do_set_memsize(self, lMemsize):
        '''
        Sets the size of the datapoints taken

        Input:
            lMemsize (int) : number of datapoints

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set memsize to %s' % lMemsize)
        self._set_param(_spcm_regs.SPC_MEMSIZE, lMemsize)

    def do_get_memsize(self):
        '''
        Get the number of datapoints that are read out

        Input:
            None

        Output:
            memsize (int) : number of datapoints
        '''
        logging.debug(__name__ + ' : Get memzise')
        return self._get_param(_spcm_regs.SPC_MEMSIZE)

### channel parameters

    def do_set_input_path_ch0(self, path):
        '''
        sets the input path of channel 0
        channel 0 impedance is 1MOhm
        channel 1 impedance is 50Ohm

        Input:
            path (int): selected path
        '''
        amp = self._get_param(_spcm_regs.SPC_AMP0)
        offset = self._get_param(_spcm_regs.SPC_OFFS0)
        self._set_param(_spcm_regs.SPC_PATH0, path)
        self._set_param(_spcm_regs.SPC_AMP0, amp)
        self._set_param(_spcm_regs.SPC_OFFS0, offset)

    def do_set_input_path_ch1(self, path):
        '''
        sets the input path of channel 1
        channel 0 impedance is 1MOhm
        channel 1 impedance is 50Ohm

        Input:
            path (int): selected path
        '''
        amp = self._get_param(_spcm_regs.SPC_AMP1)
        offset = self._get_param(_spcm_regs.SPC_OFFS1)
        self._set_param(_spcm_regs.SPC_PATH1, path)
        self._set_param(_spcm_regs.SPC_AMP1, amp)
        self._set_param(_spcm_regs.SPC_OFFS1, offset)

    def do_set_input_amp_ch0(self, amp):
        '''
        Sets the amplitude of the range of channel 0
        The range defines the precision of the analog-digital conversion

        Input:
            amp (int): amplitude of the channel in millivolts

        Output:
            None
        '''
        logging.debug(__name__ + ' : Setting input amp0 to %s' % amp)
        self._set_param(_spcm_regs.SPC_AMP0, amp)

    def do_set_input_amp_ch1(self, amp):
        '''
        Sets the amplitude of the range of channel 1
        The range defines the precision of the analog-digital conversion

        Input:
            amp (int): amplitude of the channel in millivolts

        Output:
            None
        '''
        logging.debug(__name__ + ' : Setting input amp1 to %s' % amp)
        self._set_param(_spcm_regs.SPC_AMP1, amp)

    def do_set_input_offset_ch0(self, offset):
        '''
        Sets the offset of channel 0 as a percentage
        of the range

        Input:
            offset (int): percentage of range

        Output:
            None
        '''
        logging.debug(__name__ + ' : Setting input offset0 to %s' % offset)
        self._set_param(_spcm_regs.SPC_OFFS0, offset)

    def do_set_input_offset_ch1(self, offset):
        '''
        Sets the offset of channel 1 as a percentage
        of the range

        Input:
            offset (int): percentage of range

        Output:
            None
        '''
        logging.debug(__name__ + ' : Setting input offset1 to %s' % offset)
        self._set_param(_spcm_regs.SPC_OFFS1, offset)

    def do_get_input_path_ch0(self):
        '''
        gets the input path of channel 0
        channel 0 impedance is 1MOhm
        channel 1 impedance is 50Ohm

        Output:
            path (int): selected path
        '''
        self._get_param(_spcm_regs.SPC_PATH0)

    def do_get_input_path_ch1(self):
        '''
        gets the input path of channel 1
        channel 0 impedance is 1MOhm
        channel 1 impedance is 50Ohm

        Output:
            path (int): selected path
        '''
        self._get_param(_spcm_regs.SPC_PATH1)

    def do_get_input_amp_ch0(self):
        '''
        Gets the amplitude of the range of channel 0
        The range defines the precision of the analog-digital conversion

        Input:
            None

        Output:
            amp (int): amplitude of the channel in millivolts
        '''
        logging.debug(__name__ + ' : Getting input amp0')
        return self._get_param(_spcm_regs.SPC_AMP0)

    def do_get_input_amp_ch1(self):
        '''
        Gets the amplitude of the range of channel 1
        The range defines the precision of the analog-digital conversion

        Input:
            None

        Output:
            amp (int): amplitude of the channel in millivolts
        '''
        logging.debug(__name__ + ' : Getting input amp1')
        return self._get_param(_spcm_regs.SPC_AMP1)

    def do_get_input_offset_ch0(self):
        '''
        Gets the offset of channel 0 as a percentage
        of the range

        Input:
            None

        Output:
            offset (int): percentage of range
        '''
        logging.debug(__name__ + ' : Getting input offset0')
        return self._get_param(_spcm_regs.SPC_OFFS0)

    def do_get_input_offset_ch1(self):
        '''
        Gets the offset of channel 1 as a percentage
        of the range

        Input:
            None

        Output:
            offset (int): percentage of range
        '''
        logging.debug(__name__ + ' : Getting input offset1')
        return self._get_param(_spcm_regs.SPC_OFFS1)

### clock

    def do_set_spc_samplerate(self, rate):
        '''
        defines the sampling rate in Hz for internal
        sample rate generation

        Input:
            rate (int) : sample rate in Hz, minimum is 10MSPS for M3i cards

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set spc samplerate to %s' % rate)
        rate = int(rate)
        self._set_param(_spcm_regs.SPC_SAMPLERATE, rate)

    def do_get_spc_samplerate(self):
        '''
        gets the sampling rate in Hz for internal
        sample rate generation

        Input:
            None

        Output:
            rate (int) : sample rate in Hz
        '''
        logging.debug(__name__ + ' : Get spc samplerate')
        return self._get_param(_spcm_regs.SPC_SAMPLERATE)

    def do_set_reference_clock(self, freq):
        '''
        Programs the external reference clock

        Input:
            freq (int) : frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set reference clock freq to %s' % freq)
        self._set_param(_spcm_regs.SPC_REFERENCECLOCK, freq)

    def do_get_reference_clock(self):
        '''
        Gets the external reference clock setting

        Input:
            None

        Output:
            freq (int) : frequency in Hz
        '''
        logging.debug(__name__ + ' : Get reference clock setting')
        return self._get_param(_spcm_regs.SPC_REFERENCECLOCK)
